# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Python barcode renderer for Code39 barcodes. Designed for use with Inkscape.
"""

from .Base import Barcode

ENCODE = {
    "0": "000110100",
    "1": "100100001",
    "2": "001100001",
    "3": "101100000",
    "4": "000110001",
    "5": "100110000",
    "6": "001110000",
    "7": "000100101",
    "8": "100100100",
    "9": "001100100",
    "A": "100001001",
    "B": "001001001",
    "C": "101001000",
    "D": "000011001",
    "E": "100011000",
    "F": "001011000",
    "G": "000001101",
    "H": "100001100",
    "I": "001001100",
    "J": "000011100",
    "K": "100000011",
    "L": "001000011",
    "M": "101000010",
    "N": "000010011",
    "O": "100010010",
    "P": "001010010",
    "Q": "000000111",
    "R": "100000110",
    "S": "001000110",
    "T": "000010110",
    "U": "110000001",
    "V": "011000001",
    "W": "111000000",
    "X": "010010001",
    "Y": "110010000",
    "Z": "011010000",
    "-": "010000101",
    "*": "010010100",
    "+": "010001010",
    "$": "010101000",
    "%": "000101010",
    "/": "010100010",
    ".": "110000100",
    " ": "011000100",
}


class Code39(Barcode):
    """Convert a text into string binary of black and white markers"""

    def encode(self, text):
        self.text = text.upper()
        result = ""
        # It is possible for us to encode code39
        # into full ascii, but this feature is
        # not enabled here
        for char in "*" + self.text + "*":
            if char not in ENCODE:
                char = "-"
            result = result + ENCODE[char] + "0"

        # Now we need to encode the code39, best read
        # the code to understand what it's up to:
        encoded = ""
        colour = "1"  # 1 = Black, 0 = White
        for data in result:
            if data == "1":
                encoded = encoded + colour + colour
            else:
                encoded = encoded + colour
            colour = "0" if colour == "1" else "1"

        return encoded



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Windell Oskay <windell@oskay.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=attribute-defined-outside-init
#
"""
Provide a way to load lxml attributes with an svg API on top.
"""

import random
import math
import re

from lxml import etree

from ..css import ConditionalRule
from ..interfaces.IElement import ISVGDocumentElement

from ..deprecated.meta import DeprecatedSvgMixin, deprecate
from ..units import discover_unit, parse_unit
from ._selected import ElementList
from ..transforms import BoundingBox
from ..styles import StyleSheets

from ._base import BaseElement
from ._meta import StyleElement

from typing import Optional, List

if False:  # pylint: disable=using-constant-test
    import typing  # pylint: disable=unused-import


class SvgDocumentElement(DeprecatedSvgMixin, ISVGDocumentElement, BaseElement):
    """Provide access to the document level svg functionality"""

    # pylint: disable=too-many-public-methods
    tag_name = "svg"

    selection: ElementList
    """The selection as passed by Inkscape (readonly)"""

    def _init(self):
        self.current_layer = None
        self.view_center = (0.0, 0.0)
        self.selection = ElementList(self)
        self.ids = {}

    def tostring(self):
        """Convert document to string"""
        return etree.tostring(etree.ElementTree(self))

    def get_ids(self):
        """Returns a set of unique document ids"""
        if not self.ids:
            self.ids = set(self.xpath("//@id"))
        return self.ids

    def get_unique_id(
        self,
        prefix: str,
        size: Optional[int] = None,
        blacklist: Optional[List[str]] = None,
    ):
        """Generate a new id from an existing old_id

        The id consists of a prefix and an appended random integer with size digits.

        If size is not given, it is determined automatically from the length of
        existing ids, i.e. those in the document plus those in the blacklist.

        Args:
            prefix (str): the prefix of the new ID.
            size (Optional[int], optional): number of digits of the second part of the
                id. If None, the length is chosen based on the amount of existing
                objects. Defaults to None.

                .. versionchanged:: 1.1
                    The default of this parameter has been changed from 4 to None.
            blacklist (Optional[Iterable[str]], optional): An additional iterable of ids
                that are not allowed to be used. This is useful when bulk inserting
                objects.
                Defaults to None.

                .. versionadded:: 1.2

        Returns:
            _type_: _description_
        """
        ids = self.get_ids()
        if size is None:
            size = max(math.ceil(math.log10(len(ids) or 1000)) + 1, 4)
        if blacklist is not None:
            ids.update(blacklist)
        new_id = None
        _from = 10**size - 1
        _to = 10**size
        while new_id is None or new_id in ids:
            # Do not use randint because py2/3 incompatibility
            new_id = prefix + str(int(random.random() * _from - _to) + _to)
        self.ids.add(new_id)
        return new_id

    def get_page_bbox(self):
        """Gets the page dimensions as a bbox"""
        return BoundingBox(
            (0, float(self.viewbox_width)), (0, float(self.viewbox_height))
        )

    def get_current_layer(self):
        """Returns the currently selected layer"""
        layer = self.getElementById(self.namedview.current_layer, "svg:g")
        if layer is None:
            return self
        return layer

    def getElement(self, xpath):  # pylint: disable=invalid-name
        """Gets a single element from the given xpath or returns None"""
        return self.findone(xpath)

    def getElementById(
        self, eid: str, elm="*", literal=False
    ):  # pylint: disable=invalid-name
        """Get an element in this svg document by it's ID attribute.

        Args:
            eid (str): element id
            elm (str, optional): element type, including namespace, e.g. ``svg:path``.
                Defaults to "*".
            literal (bool, optional): If ``False``, ``#url()`` is stripped from ``eid``.
                Defaults to False.

                .. versionadded:: 1.1

        Returns:
            Union[BaseElement, None]: found element
        """
        if eid is not None and not literal:
            eid = eid.strip()[4:-1] if eid.startswith("url(") else eid
            eid = eid.lstrip("#")
        return self.getElement(f'//{elm}[@id="{eid}"]')

    def getElementByName(self, name, elm="*"):  # pylint: disable=invalid-name
        """Get an element by it's inkscape:label (aka name)"""
        return self.getElement(f'//{elm}[@inkscape:label="{name}"]')

    def getElementsByClass(self, class_name):  # pylint: disable=invalid-name
        """Get elements by it's class name"""

        return self.xpath(ConditionalRule(f".{class_name}").to_xpath())

    def getElementsByHref(
        self, eid: str, attribute="xlink:href"
    ):  # pylint: disable=invalid-name
        """Get elements that reference the element with id eid.

        Args:
            eid (str): _description_
            attribute (str, optional): Attribute to look for.
                Valid choices: "xlink:href", "mask", "clip-path".
                Defaults to "xlink:href".

                .. versionadded:: 1.2

        Returns:
            Any: list of elements
        """
        if attribute == "xlink:href":
            return self.xpath(f'//*[@xlink:href="#{eid}"]')
        elif attribute == "mask":
            return self.xpath(f'//*[@mask="url(#{eid})"]')
        elif attribute == "clip-path":
            return self.xpath(f'//*[@clip-path="url(#{eid})"]')

    def getElementsByStyleUrl(self, eid, style=None):  # pylint: disable=invalid-name
        """Get elements by a style attribute url"""
        url = f"url(#{eid})"
        if style is not None:
            url = style + ":" + url
        return self.xpath(f'//*[contains(@style,"{url}")]')

    @property
    def name(self):
        """Returns the Document Name"""
        return self.get("sodipodi:docname", "")

    @property
    def namedview(self):
        """Return the sp namedview meta information element"""
        return self.get_or_create("//sodipodi:namedview", prepend=True)

    @property
    def metadata(self):
        """Return the svg metadata meta element container"""
        return self.get_or_create("//svg:metadata", prepend=True)

    @property
    def defs(self):
        """Return the svg defs meta element container"""
        return self.get_or_create("//svg:defs", prepend=True)

    def get_viewbox(self):
        """Parse and return the document's viewBox attribute"""
        try:
            ret = [
                float(unit) for unit in re.split(r",\s*|\s+", self.get("viewBox", "0"))
            ]
        except ValueError:
            ret = ""
        if len(ret) != 4:
            return [0, 0, 0, 0]
        return ret

    @property
    def viewbox_width(self) -> float:  # getDocumentWidth(self):
        """Returns the width of the `user coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e.
        the width of the viewbox, as defined in the SVG file. If no viewbox is defined,
        the value of the width attribute is returned. If the height is not defined,
        returns 0.

        .. versionadded:: 1.2"""
        return self.get_viewbox()[2] or self.viewport_width

    @property
    def viewport_width(self) -> float:
        """Returns the width of the `viewport coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        width attribute of the svg element converted to px

        .. versionadded:: 1.2"""
        return self.to_dimensionless(self.get("width")) or self.get_viewbox()[2]

    @property
    def viewbox_height(self) -> float:  # getDocumentHeight(self):
        """Returns the height of the `user coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        height of the viewbox, as defined in the SVG file. If no viewbox is defined, the
        value of the height attribute is returned. If the height is not defined,
        returns 0.

        .. versionadded:: 1.2"""
        return self.get_viewbox()[3] or self.viewport_height

    @property
    def viewport_height(self) -> float:
        """Returns the width of the `viewport coordinate system
        <https://www.w3.org/TR/SVG2/coords.html#Introduction>`_ in user units, i.e. the
        height attribute of the svg element converted to px

        .. versionadded:: 1.2"""
        return self.to_dimensionless(self.get("height")) or self.get_viewbox()[3]

    @property
    def scale(self):
        """Returns the ratio between the viewBox width and the page width.

        .. versionchanged:: 1.2
            Previously, the scale as shown by the document properties was computed,
            but the computation of this in core Inkscape changed in Inkscape 1.2, so
            this was moved to :attr:`inkscape_scale`."""
        return self._base_scale()

    @property
    def inkscape_scale(self):
        """Returns the ratio between the viewBox width (in width/height units) and the
        page width, which is displayed as "scale" in the Inkscape document
        properties.

        .. versionadded:: 1.2"""

        viewbox_unit = (
            parse_unit(self.get("width")) or parse_unit(self.get("height")) or (0, "px")
        )[1]
        return self._base_scale(viewbox_unit)

    def _base_scale(self, unit="px"):
        """Returns what Inkscape shows as "user units per `unit`"

        .. versionadded:: 1.2"""
        try:
            scale_x = (
                self.to_dimensional(self.viewport_width, unit) / self.viewbox_width
            )
            scale_y = (
                self.to_dimensional(self.viewport_height, unit) / self.viewbox_height
            )
            value = max([scale_x, scale_y])
            return 1.0 if value == 0 else value
        except (ValueError, ZeroDivisionError):
            return 1.0

    @property
    def equivalent_transform_scale(self) -> float:
        """Return the scale of the equivalent transform of the svg tag, as defined by
        https://www.w3.org/TR/SVG2/coords.html#ComputingAViewportsTransform
        (highly simplified)

        .. versionadded:: 1.2"""
        return self.scale

    @property
    def unit(self):
        """Returns the unit used for in the SVG document.
        In the case the SVG document lacks an attribute that explicitly
        defines what units are used for SVG coordinates, it tries to calculate
        the unit from the SVG width and viewBox attributes.
        Defaults to 'px' units."""
        if not hasattr(self, "_unit"):
            self._unit = "px"  # Default is px
            viewbox = self.get_viewbox()
            if viewbox and set(viewbox) != {0}:
                self._unit = discover_unit(self.get("width"), viewbox[2], default="px")
        return self._unit

    @property
    def document_unit(self):
        """Returns the display unit (Inkscape-specific attribute) of the document

        .. versionadded:: 1.2"""
        return self.namedview.get("inkscape:document-units", "px")

    @property
    def stylesheets(self):
        """Get all the stylesheets, bound together to one, (for reading)"""
        sheets = StyleSheets(self)
        for node in self.xpath("//svg:style"):
            sheets.append(node.stylesheet())
        return sheets

    @property
    def stylesheet(self):
        """Return the first stylesheet or create one if needed (for writing)"""
        for sheet in self.stylesheets:
            return sheet

        style_node = StyleElement()
        self.defs.append(style_node)
        return style_node.stylesheet()


def width(self):
    """Use :func:`viewport_width` instead"""
    return self.viewport_width


def height(self):
    """Use :func:`viewport_height` instead"""
    return self.viewport_height


SvgDocumentElement.width = property(deprecate(width, "1.2"))
SvgDocumentElement.height = property(deprecate(height, "1.2"))



# coding=utf-8
#
# Copyright (C) 2016 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110, USA.
#
"""
Python barcode renderer for EAN2 barcodes. Designed for use with Inkscape.
"""

from .BaseEan import EanBarcode

FAMS = ["00", "01", "10", "11"]
START = "01011"


class Ean2(EanBarcode):
    """Provide an Ean5 barcode generator"""

    length = 2
    name = "ean5"

    def _encode(self, num, guide=False):
        if len(num) != 2:
            num = ([0, 0] + num)[-2:]
        self.text = " ".join(EanBarcode.space(num))
        family = ((num[0] * 10) + num[1]) % 4
        return START + "01".join(EanBarcode.encode_interleaved(family, num, FAMS))



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Inkscape's general barcode extension. Run from within inkscape or use the
Barcode module provided for outside or scripting.
"""

import inkex
from barcode import get_barcode


class Barcode(inkex.GenerateExtension):
    """
    Raw barcode Effect class, see Barcode base class.
    """

    def add_arguments(self, pars):
        pars.add_argument("-l", "--height", type=int, default=30, help="Barcode Height")
        pars.add_argument("-t", "--type", default="", help="Barcode Type")
        pars.add_argument("-d", "--text", default="", help="Text to print on barcode")

    def generate(self):
        (pos_x, pos_y) = self.svg.namedview.center

        return get_barcode(
            self.options.type,
            text=self.options.text,
            height=self.options.height,
            document=self.document,
            x=pos_x,
            y=pos_y,
            scale=self.svg.unittouu("1px"),
        ).generate()


if __name__ == "__main__":
    Barcode().run()



# coding=utf-8
#
# Copyright 2011-2022 Martin Owens <doctormo@geek-2.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Gtk Application base classes, providing a way to load a GtkBuilder
with a specific glade/ui file conaining windows, and building
a usable pythonic interface from them.
"""
import os
import signal
import logging

from gi.repository import Gtk, GLib


class GtkApp:
    """
    This wraps gtk builder and allows for some extra functionality with
    windows, especially the management of gtk main loops.

    Args:
        start_loop (bool, optional): If set to true will start a new gtk main loop.
            Defaults to False.
        start_gui (bool, optional): Used as local propertes if unset and passed to
                primary window when loaded. Defaults to True.
    """

    @property
    def prefix(self):
        """Folder prefix added to ui_dir"""
        return self.kwargs.get("prefix", "")

    @property
    def windows(self):
        """Returns a list of windows for this app"""
        return self.kwargs.get("windows", [])

    @property
    def ui_dir(self):
        """This is often the local directory"""
        return self.kwargs.get("ui_dir", "./")

    @property
    def ui_file(self):
        """If a single file is used for multiple windows"""
        return self.kwargs.get("ui_file", None)

    @property
    def app_name(self):
        """Set this variable in your class"""
        try:
            return self.kwargs["app_name"]
        except KeyError:
            raise NotImplementedError(
                "App name is not set, pass in or set 'app_name' in class."
            )

    @property
    def window(self):
        """Return the primary window"""
        return self._primary

    def __init__(self, start_loop=False, start_gui=True, **kwargs):
        """Creates a new GtkApp."""
        self.kwargs = kwargs
        self._loaded = {}
        self._initial = {}
        self._primary = None

        self.main_loop = GLib.main_depth()

        # Start with creating all the defined windows.
        if start_gui:
            self.init_gui()
        # Start up a gtk main loop when requested
        if start_loop:
            self.run()

    def run(self):
        """Run the gtk mainloop with ctrl+C and keyboard interupt additions"""
        if not Gtk.init_check()[0]:  # pragma: no cover
            raise RuntimeError(
                "Gtk failed to start." " Make sure $DISPLAY variable is set.\n"
            )
        try:
            # Add a signal to force quit on Ctrl+C (just like the old days)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            Gtk.main()
        except KeyboardInterrupt:  # pragma: no cover
            logging.info("User Interputed")
        logging.debug("Exiting %s", self.app_name)

    def get_ui_file(self, window):
        """Load any given gtk builder file from a standard location."""
        paths = [
            os.path.join(self.ui_dir, self.prefix, f"{window}.ui"),
            os.path.join(self.ui_dir, self.prefix, f"{self.ui_file}.ui"),
        ]
        for path in paths:
            if os.path.isfile(path):
                return path
        raise FileNotFoundError(f"Gtk Builder file is missing: {paths}")

    def init_gui(self):
        """Initalise all of our windows and load their signals"""
        if self.windows:
            for cls in self.windows:
                window = cls
                logging.debug("Adding window %s to GtkApp", window.name)
                self._initial[window.name] = window
            for window in self._initial.values():
                if window.primary:
                    if not self._primary:
                        self._primary = self.load_window(window.name)
        if not self.windows or not self._primary:
            raise KeyError(f"No primary window found for '{self.app_name}' app.")

    def load_window(self, name, *args, **kwargs):
        """Load a specific window from our group of windows"""
        window = self.proto_window(name)
        window.init(*args, **kwargs)
        return window

    def load_window_extract(self, name, **kwargs):
        """Load a child window as a widget container"""
        window = self.proto_window(name)
        window.load_widgets(**kwargs)
        return window.extract()

    def proto_window(self, name):
        """
        Loads a glade window as a window without initialisation, used for
        extracting widgets from windows without loading them as windows.
        """
        logging.debug("Loading '%s' from %s", name, self._initial)
        if name in self._initial:
            # Create a new instance of this window
            window = self._initial[name](self)
            # Save the window object linked against the gtk window instance
            self._loaded[window.wid] = window
            return window
        raise KeyError(f"Can't load window '{name}', class not found.")

    def remove_window(self, window):
        """Remove the window from the list and exit if none remain"""
        if window.wid in self._loaded:
            self._loaded.pop(window.wid)
        else:
            logging.warning("Missing window '%s' on exit.", window.name)
        logging.debug("Loaded windows: %s", self._loaded)
        if not self._loaded:
            self.exit()

    def exit(self):
        """Exit our gtk application and kill gtk main if we have to"""
        if self.main_loop < GLib.main_depth():
            # Quit Gtk loop if we started one.
            tag = self._primary.name if self._primary else "program"
            logging.debug("Quit '%s' Main Loop.", tag)
            Gtk.main_quit()
        # You have to return in order for the loop to exit
        return 0



# coding=utf-8
#
# Copyright (c) 2018 - Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
The ultimate base functionality for every Inkscape extension.
"""

import os
import re
import sys
import copy

from typing import (
    Dict,
    List,
    Tuple,
    Type,
    Optional,
    Callable,
    Any,
    Union,
    IO,
    TYPE_CHECKING,
    cast,
)

from argparse import ArgumentParser, Namespace
from lxml import etree

from .interfaces.IElement import IBaseElement, ISVGDocumentElement
from .utils import filename_arg, AbortExtension, ABORT_STATUS, errormsg, do_nothing
from .elements._parser import load_svg
from .elements._utils import NSS
from .localization import localize

stdout = sys.stdout.buffer  # type: ignore


class InkscapeExtension:
    """
    The base class extension, provides argument parsing and basic
    variable handling features.
    """

    multi_inx = False  # Set to true if this class is used by multiple inx files.
    extra_nss = {}  # type: Dict[str, str]

    def __init__(self):
        # type: () -> None
        NSS.update(self.extra_nss)
        self.file_io = None  # type: Optional[IO]
        self.options = Namespace()
        self.document = None  # type: Union[None, bytes, str, etree.element]
        self.arg_parser = ArgumentParser(description=self.__doc__)

        self.arg_parser.add_argument(
            "input_file",
            nargs="?",
            metavar="INPUT_FILE",
            type=filename_arg,
            help="Filename of the input file (default is stdin)",
            default=None,
        )

        self.arg_parser.add_argument(
            "--output",
            type=str,
            default=None,
            help="Optional output filename for saving the result (default is stdout).",
        )

        self.add_arguments(self.arg_parser)

        localize()

    def add_arguments(self, pars):
        # type: (ArgumentParser) -> None
        """Add any extra arguments to your extension handle, use:

        def add_arguments(self, pars):
            pars.add_argument("--num-cool-things", type=int, default=3)
            pars.add_argument("--pos-in-doc", type=str, default="doobry")
        """
        # No extra arguments by default so super is not required

    def parse_arguments(self, args):
        # type: (List[str]) -> None
        """Parse the given arguments and set 'self.options'"""
        self.options = self.arg_parser.parse_args(args)

    def arg_method(self, prefix="method"):
        # type: (str) -> Callable[[str], Callable[[Any], Any]]
        """Used by add_argument to match a tab selection with an object method

        pars.add_argument("--tab", type=self.arg_method(), default="foo")
        ...
        self.options.tab(arguments)
        ...
        .. code-block:: python
        .. def method_foo(self, arguments):
        ..     # do something
        """

        def _inner(value):
            name = f"""{prefix}_{value.strip('"').lower()}""".replace("-", "_")
            try:
                return getattr(self, name)
            except AttributeError as error:
                if name.startswith("_"):
                    return do_nothing
                raise AbortExtension(f"Can not find method {name}") from error

        return _inner

    @staticmethod
    def arg_number_ranges():

        """Parses a number descriptor. e.g:
        ``1,2,4-5,7,9-`` is parsed to ``1, 2, 4, 5, 7, 9, 10, ..., lastvalue``

        .. versionadded:: 1.2

        Usage:

        .. code-block:: python

            # in add_arguments()
            pars.add_argument("--pages", type=self.arg_number_ranges(), default=1-)
            # later on, pages is then a list of ints
            pages = self.options.pages(lastvalue)

        """

        def _inner(value):
            def method(pages, lastvalue, startvalue=1):
                # replace ranges, such as -3, 10- with startvalue,2,3,10..lastvalue
                pages = re.sub(
                    r"(\d+|)\s?-\s?(\d+|)",
                    lambda m: ",".join(
                        map(
                            str,
                            range(
                                int(m.group(1) or startvalue),
                                int(m.group(2) or lastvalue) + 1,
                            ),
                        )
                    )
                    if not (m.group(1) or m.group(2)) == ""
                    else "",
                    pages,
                )

                pages = map(int, re.findall(r"(\d+)", pages))
                pages = tuple({i for i in pages if i <= lastvalue})
                return pages

            return lambda lastvalue, startvalue=1: method(
                value, lastvalue, startvalue=startvalue
            )

        return _inner

    @staticmethod
    def arg_class(options: List[Type]) -> Callable[[str], Any]:
        """Used by add_argument to match an option with a class

        Types to choose from are given by the options list

        .. versionadded:: 1.2

        Usage:

        .. code-block:: python

            pars.add_argument("--class", type=self.arg_class([ClassA, ClassB]),
                              default="ClassA")
        """

        def _inner(value: str):
            name = value.strip('"')
            for i in options:
                if name == i.__name__:
                    return i
            raise AbortExtension(f"Can not find class {name}")

        return _inner

    def debug(self, msg):
        # type: (str) -> None
        """Write a debug message"""
        errormsg(f"DEBUG<{type(self).__name__}> {msg}\n")

    @staticmethod
    def msg(msg):
        # type: (str) -> None
        """Write a non-error message"""
        errormsg(msg)

    def run(self, args=None, output=stdout):
        # type: (Optional[List[str]], Union[str, IO]) -> None
        """Main entrypoint for any Inkscape Extension"""
        try:
            if args is None:
                args = sys.argv[1:]

            self.parse_arguments(args)
            if self.options.input_file is None:
                self.options.input_file = sys.stdin
            elif "DOCUMENT_PATH" not in os.environ:
                os.environ["DOCUMENT_PATH"] = self.options.input_file

            if self.options.output is None:
                self.options.output = output

            self.load_raw()
            self.save_raw(self.effect())
        except AbortExtension as err:
            errormsg(str(err))
            sys.exit(ABORT_STATUS)
        finally:
            self.clean_up()

    def load_raw(self):
        # type: () -> None
        """Load the input stream or filename, save everything to self"""
        if isinstance(self.options.input_file, str):
            # pylint: disable=consider-using-with
            self.file_io = open(self.options.input_file, "rb")
            document = self.load(self.file_io)
        else:
            document = self.load(self.options.input_file)
        self.document = document

    def save_raw(self, ret):
        # type: (Any) -> None
        """Save to the output stream, use everything from self"""
        if self.has_changed(ret):
            if isinstance(self.options.output, str):
                with open(self.options.output, "wb") as stream:
                    self.save(stream)
            else:
                if sys.platform == "win32" and not "PYTEST_CURRENT_TEST" in os.environ:
                    # When calling an extension from within Inkscape on Windows,
                    # Python thinks that the output stream is seekable
                    # (https://gitlab.com/inkscape/inkscape/-/issues/3273)
                    self.options.output.seekable = lambda self: False

                    def seek_replacement(offset: int, whence: int = 0):
                        raise AttributeError(
                            "We can't seek in the stream passed by Inkscape on Windows"
                        )

                    def tell_replacement():
                        raise AttributeError(
                            "We can't tell in the stream passed by Inkscape on Windows"
                        )

                    # Some libraries (e.g. ZipFile) don't query seekable, but check for an error
                    # on seek
                    self.options.output.seek = seek_replacement
                    self.options.output.tell = tell_replacement
                self.save(self.options.output)

    def load(self, stream):
        # type: (IO) -> str
        """Takes the input stream and creates a document for parsing"""
        raise NotImplementedError(f"No input handle for {self.name}")

    def save(self, stream):
        # type: (IO) -> None
        """Save the given document to the output file"""
        raise NotImplementedError(f"No output handle for {self.name}")

    def effect(self):
        # type: () -> Any
        """Apply some effects on the document or local context"""
        raise NotImplementedError(f"No effect handle for {self.name}")

    def has_changed(self, ret):  # pylint: disable=no-self-use
        # type: (Any) -> bool
        """Return true if the output should be saved"""
        return ret is not False

    def clean_up(self):
        # type: () -> None
        """Clean up any open handles and other items"""
        if self.file_io is not None:
            self.file_io.close()

    @classmethod
    def svg_path(cls, default=None):
        # type: (Optional[str]) -> Optional[str]
        """
        Return the folder the svg is contained in.
        Returns None if there is no file.

        .. versionchanged:: 1.1
            A default path can be given which is returned in case no path to the
            SVG file can be determined.
        """
        path = cls.document_path()
        if path:
            return os.path.dirname(path)
        if default:
            return default
        return path  # Return None or '' for context

    @classmethod
    def ext_path(cls):
        # type: () -> str
        """Return the folder the extension script is in"""
        return os.path.dirname(sys.modules[cls.__module__].__file__ or "")

    @classmethod
    def get_resource(cls, name, abort_on_fail=True):
        # type: (str, bool) -> str
        """Return the full filename of the resource in the extension's dir

        .. versionadded:: 1.1"""
        filename = cls.absolute_href(name, cwd=cls.ext_path())
        if abort_on_fail and not os.path.isfile(filename):
            raise AbortExtension(f"Could not find resource file: {filename}")
        return filename

    @classmethod
    def document_path(cls):
        # type: () -> Optional[str]
        """Returns the saved location of the document

         * Normal return is a string containing the saved location
         * Empty string means the document was never saved
         * 'None' means this version of Inkscape doesn't support DOCUMENT_PATH

        DO NOT READ OR WRITE TO THE DOCUMENT FILENAME!

         * Inkscape may have not written the latest changes, leaving you reading old
           data.
         * Inkscape will not respect anything you write to the file, causing data loss.

        .. versionadded:: 1.1
        """
        return os.environ.get("DOCUMENT_PATH", None)

    @classmethod
    def absolute_href(cls, filename, default="~/", cwd=None):
        # type: (str, str, Optional[str]) -> str
        """
        Process the filename such that it's turned into an absolute filename
        with the working directory being the directory of the loaded svg.

        User's home folder is also resolved. So '~/a.png` will be `/home/bob/a.png`

        Default is a fallback working directory to use if the svg's filename is not
        available.

        .. versionchanged:: 1.1
            If you set default to None, then the user will be given errors if
            there's no working directory available from Inkscape.
        """
        filename = os.path.expanduser(filename)
        if not os.path.isabs(filename):
            filename = os.path.expanduser(filename)
        if not os.path.isabs(filename):
            if cwd is None:
                cwd = cls.svg_path(default)
                if cwd is None:
                    raise AbortExtension(
                        "Can not use relative path, Inkscape isn't telling us the "
                        "current working directory."
                    )
                if cwd == "":
                    raise AbortExtension(
                        "The SVG must be saved before you can use relative paths."
                    )
            filename = os.path.join(cwd, filename)
        return os.path.realpath(os.path.expanduser(filename))

    @property
    def name(self):
        # type: () -> str
        """Return a fixed name for this extension"""
        return type(self).__name__


if TYPE_CHECKING:
    _Base = InkscapeExtension
else:
    _Base = object


class TempDirMixin(_Base):  # pylint: disable=abstract-method
    """
    Provide a temporary directory for extensions to stash files.
    """

    dir_suffix = ""
    dir_prefix = "inktmp"

    def __init__(self, *args, **kwargs):
        self.tempdir = None
        super().__init__(*args, **kwargs)

    def load_raw(self):
        # type: () -> None
        """Create the temporary directory"""
        # pylint: disable=import-outside-toplevel
        from tempfile import TemporaryDirectory

        # Need to hold a reference to the Directory object or else it might get GC'd
        self._tempdir = TemporaryDirectory(  # pylint: disable=consider-using-with
            prefix=self.dir_prefix, suffix=self.dir_suffix
        )
        self.tempdir = os.path.realpath(self._tempdir.name)
        super().load_raw()

    def clean_up(self):
        # type: () -> None
        """Delete the temporary directory"""
        self.tempdir = None
        self._tempdir.cleanup()
        super().clean_up()


class SvgInputMixin(_Base):  # pylint: disable=too-few-public-methods, abstract-method
    """
    Expects the file input to be an svg document and will parse it.
    """

    # Select all objects if none are selected
    select_all: Tuple[Type[IBaseElement], ...] = ()

    def __init__(self):
        super().__init__()

        self.arg_parser.add_argument(
            "--id",
            action="append",
            type=str,
            dest="ids",
            default=[],
            help="id attribute of object to manipulate",
        )

        self.arg_parser.add_argument(
            "--selected-nodes",
            action="append",
            type=str,
            dest="selected_nodes",
            default=[],
            help="id:subpath:position of selected nodes, if any",
        )

    def load(self, stream):
        # type: (IO) -> etree
        """Load the stream as an svg xml etree and make a backup"""
        document = load_svg(stream)
        self.original_document = copy.deepcopy(document)
        self.svg: ISVGDocumentElement = document.getroot()
        self.svg.selection.set(*self.options.ids)
        if not self.svg.selection and self.select_all:
            self.svg.selection = self.svg.descendants().filter(*self.select_all)
        return document


class SvgOutputMixin(_Base):  # pylint: disable=too-few-public-methods, abstract-method
    """
    Expects the output document to be an svg document and will write an etree xml.

    A template can be specified to kick off the svg document building process.
    """

    template = """<svg viewBox="0 0 {width} {height}"
        width="{width}{unit}" height="{height}{unit}"
        xmlns="http://www.w3.org/2000/svg" xmlns:svg="http://www.w3.org/2000/svg"
        xmlns:xlink="http://www.w3.org/1999/xlink"
        xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
        xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape">
    </svg>"""

    @classmethod
    def get_template(cls, **kwargs):
        """
        Opens a template svg document for building, the kwargs
        MUST include all the replacement values in the template, the
        default template has 'width' and 'height' of the document.
        """
        kwargs.setdefault("unit", "")
        return load_svg(str(cls.template.format(**kwargs)))

    def save(self, stream):
        # type: (IO) -> None
        """Save the svg document to the given stream"""
        if isinstance(self.document, (bytes, str)):
            document = self.document
        elif "Element" in type(self.document).__name__:
            # isinstance can't be used here because etree is broken
            doc = cast(etree, self.document)
            document = doc.getroot().tostring()
        else:
            raise ValueError(
                f"Unknown type of document: {type(self.document).__name__} can not"
                + "save."
            )

        try:
            stream.write(document)
        except TypeError:
            # we hope that this happens only when document needs to be encoded
            stream.write(document.encode("utf-8"))  # type: ignore


class SvgThroughMixin(SvgInputMixin, SvgOutputMixin):  # pylint: disable=abstract-method
    """
    Combine the input and output svg document handling (usually for effects).
    """

    def has_changed(self, ret):  # pylint: disable=unused-argument
        # type: (Any) -> bool
        """Return true if the svg document has changed"""
        original = etree.tostring(self.original_document)
        result = etree.tostring(self.document)
        return original != result



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Access Reactome Content Services.
"""

import re

from import_sources import RemoteSource

TAG_REX = re.compile(r'<[^<]+?>')

class Reactome(RemoteSource):
    name = 'Reactome (Bio)'
    icon = 'sources/reactome.svg'
    search_url = "https://reactome.org/ContentService/search/query"
    file_url = "https://reactome.org/icon/{stId}.svg"
    icon_url = "https://reactome.org/icon/{stId}.png"
    all_licence = "cc-by-sa-4.0"

    def search(self, query):
        params = {
            "query": query,
            "types": "Icon",
            "cluster": "true",
            "Start row": 0,
            "rows": 100,
        }
        response = {}
        try:
            response = self.session.get(self.search_url, params=params).json()
        except Exception:
            pass

        if 'messages' in response and 'No entries' in response['messages'][0]:
            return
        for cats in response.get('results', []):
            for entry in cats['entries']:
                yield {
                    'id': entry['dbId'],
                    'name': TAG_REX.sub('', entry['name']),
                    'author': 'Reactome/'+entry.get('iconDesignerName', "Unknown"),
                    'summary': TAG_REX.sub('', entry.get('summation', '')),
                    'created': None, # No data
                    'popularity': 0, # No data
                    'thumbnail': self.icon_url.format(**entry),
                    'file': self.file_url.format(**entry),
                    'license': self.all_licence,
                }




#!/usr/bin/env python
"""Convert the text to braille text"""

import inkex

# https://en.wikipedia.org/wiki/Braille_ASCII#Braille_ASCII_values
U2800_MAP = " A1B'K2L@CIF/MSP\"E3H9O6R^DJG>NTQ,*5<-U8V.%[$+X!&;:4\\0Z7(_?W]#Y)="


class Braille(inkex.TextExtension):
    """Convert to ASCII Braille"""

    @staticmethod
    def map_char(char):
        """Map a single letter to braille"""
        assert isinstance(char, str)
        try:
            mapint = U2800_MAP.index(char.upper())
        except ValueError:
            return char
        return chr(mapint + 0x2800)


if __name__ == "__main__":
    Braille().run()



#
# Copyright 2018-2022 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Information about a package"""

import os
import sys

from inkex.gui import ChildWindow, TreeView


class ExtensionTreeItem(object):
    """Shows the name of the item in the extensions tree"""

    def __init__(self, name, kind="debug", parent=None):
        self.kind = kind
        self.name = str(name)


class ExtensionTreeView(TreeView):
    """A list of extensions (inx file based)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parents = {}
        self.menus = {}

    def setup(self, *args, **kwargs):
        self.ViewColumn("Name", expand=True, text="name")
        self.ViewSort(data=lambda item: item.name, ascending=True)

    def get_menu(self, parent, *remain):
        menu_id = "::".join([str(r) for r in remain])
        if menu_id in self.menus:
            return self.menus[menu_id]

        if remain[:-1]:
            parent = self.get_menu(parent, *remain[:-1])

        menu = self._add_item([ExtensionTreeItem(remain[-1])], parent=parent)
        self.menus[menu_id] = menu
        return menu

    def add_item(self, item, parent=None):
        if not item or not item.name:
            return None
        if item.kind not in self.parents:
            tree_item = ExtensionTreeItem(item.kind.title())
            self.parents[item.kind] = self._add_item([tree_item], parent=None)
        parent = self.parents[item.kind]
        if item.kind == "effect" and len(item.menu) > 1:
            parent = self.get_menu(parent, *item.menu[:-1])
        return self._add_item([item], parent)


class MoreInformation(ChildWindow):
    """Show further information for an installed package"""

    name = "info"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inx = ExtensionTreeView(self.widget("inx"), selected=self.select_inx)

    def load_widgets(self, pixmaps, item):
        """Initialise the information"""
        try:
            self.widget("info_website").show()
            self.widget("info_website").set_uri(item.link)
            self.widget("info_website").set_label(item.link)
        except Exception:
            self.widget("info_website").hide()

        self.pixmaps = pixmaps
        self.window.set_title(item.name)
        self.widget("info_name").set_label(item.name)
        self.widget("info_desc").set_label(item.summary)
        self.widget("info_version").set_label(item.version)
        self.widget("info_license").set_label(item.license)
        self.widget("info_author").set_label(f"{item.author}")
        try:
            self.widget("info_image").set_from_pixbuf(pixmaps.get(item.get_icon()))
        except Exception:
            pass

        self.inx.clear()
        if hasattr(item, "get_files"):
            for fn in item.get_files(filters=("*.inx",)):
                self.inx.add([ExtensionTreeItem(fn, kind="lost-and-found")])

    def select_inx(self, item):
        pass



#
# Copyright 2020 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Extract information from an inx file and follow all the files.
"""

import os
import sys

from datetime import datetime
from modulefinder import ModuleFinder
from collections import defaultdict

from inkex.inx import InxFile

from inkman import __pkgname__, __version__
from inkman.utils import DATA_DIR

MODULES = (
    (
        "Depricated",
        "Using depricated modules {names}",
        "Critical",
        (
            "bezmisc",
            "cubicsuperpath",
            "simplestyle",
            "cspsubdiv",
            "ffgeom",
            "simplepath",
            "simpletransform",
        ),
    ),
    ("Unusual", "Unusual modules being used {names}", "Warning", ("argparse",)),
    ("Ignored", None, None, ("os", "sys", "lxml")),
)


class PackageInxFile(InxFile):
    def __init__(self, filename):
        super().__init__(filename)
        self.inx_file = filename

    def get_file_root(self):
        return os.path.dirname(self.inx_file)

    def get_script(self):
        """Return the file the INX points to"""
        if self.script.get("location", None) in ("inx", None):
            return os.path.join(self.get_file_root(), self.script["script"])
        raise IOError("Can't find script filename")


class IncludeFile(object):
    """A file to include in the package"""

    def __init__(self, root, filepath, target=None):
        self.root = root
        self.filepath = filepath
        self.name = os.path.basename(filepath)
        self.target = target or os.path.dirname(filepath)

    def __repr__(self):
        return f"<File name='{self.name}'>"

    def file_icon(self):
        """Return a filename type"""
        ext = self.filepath.rsplit(".", 1)[-1]
        if ext in ("inx", "py", "txt", "svg", "png"):
            return ext
        return "default"

    def detect_deps(self, modules=False):
        if not self.filepath.endswith(".py"):
            return [], []
        # We want to find all modules imported by the script directly
        # So we set the location to nowhere and collect the badmodules
        finder = ModuleFinder("never-land")
        finder.run_script(self.filepath)
        # TODO: split out local modules, system modules and genuine deps

        deps, mods = [], []
        for key, locs in finder.badmodules.items():
            if "__main__" not in locs:
                continue
            try:
                mod_path = finder.find_module("inkman", path=sys.path)[1]
                if "/python" in mod_path:
                    if "site-packages" in mod_path:
                        deps.append(key.split(".")[0])
                elif mod_path.startswith(self.root):
                    mods.append(mod_path)
            except ImportError:
                continue
        return mods, deps


class GeneratePackage(object):
    """A generated package based on data"""

    def __init__(self, inx_files, template=os.path.join(DATA_DIR, "setup.template")):
        with open(template, "r") as fhl:
            self._template = fhl.read()
        self.files = []
        self.requires = set()
        self.warnings = defaultdict(list)
        self.name = ""
        self.ident = ""

        for x, inx_file in enumerate(inx_files):
            try:
                inx = PackageInxFile(inx_file)
            except Exception as err:
                self.add_warning(f"Can't add file {inx_file}: {err}", 5)
                continue
            self._add_file(inx.get_file_root(), inx.inx_file)
            self._add_file(inx.get_file_root(), inx.get_script())
            if not x:
                self.name = inx.name or ""
                self.ident = inx.ident.replace(".", "-")

    def _add_file(self, root, fname):
        included = IncludeFile(root, fname)
        self.files.append(included)
        (mods, deps) = included.detect_deps()
        for dep in deps:
            self._add_dep(dep)
        for fname in mods:
            self._add_file(root, fname)
        print(f"DEP: {self.requires}")

    def _add_dep(self, name):
        """Try and find the package associated with this name"""
        for (list_name, warning, level, mod_list) in MODULES:
            if name in mod_list:
                self.warnings[list_name].append(name)
                return
        self.requires.add(name)

    def generate_setup(self, target):
        """Generates a setup.py file contents"""
        args = {
            "__pkgname__": __pkgname__,
            "__version__": __version__,
            "year": datetime.now().year,
            "readme": self.generate_readme(target),
            "ident": self.generate_ident(),
            "version": self.generate_version(),
            "description": self.widget("entry_name").get_text(),
            "author": self.widget("entry_author").get_text(),
            "author_email": self.widget("entry_email").get_text(),
            "url": self.widget("entry_url").get_text(),
            "license": self.widget("licenses").get_active_id(),
            "datafiles": self.generate_datafiles(),
            "requires": self.generate_requires(),
            "classifiers": "",  # TODO
        }
        return self_template.format(**args)

    def generate_ident(self):
        """Generate the ident from the text entry"""
        ident = self.widget("entry_ident").get_text()
        ident = ident.replace(" ", "-").lower()
        self.widget("entry_ident").set_text(ident)
        return "inx-" + ident

    def generate_datafiles(self):
        """Generate a list of datafile and their install location"""
        result = ""
        template = "        ('{target}', ['{file}']),\n"
        # 2. Loop through all the file objects in the list
        # 3. Add them to the final result
        return result.strip()

    def generate_requires(self):
        """Generate a list of required modules"""
        # TODO: Version keeping
        result = ""
        # 1. Loop through all the dep objects in the list
        # 2. Append them to the result
        return result


if __name__ == "__main__":
    gen = GeneratePackage(sys.argv[1:])



# coding=utf-8
#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Python barcode renderer for EAN8 barcodes. Designed for use with Inkscape.
"""

from .BaseEan import EanBarcode


class Ean8(EanBarcode):
    """Provide an EAN8 barcode generator"""

    name = "ean8"
    checks = [8]
    lengths = [7]

    def _encode(self, num, guide=False):
        """Encode an ean8 barcode"""
        self.text = EanBarcode.space(num[:4], 3, num[4:])
        return self.enclose(
            EanBarcode.encode_left(num[:4]), EanBarcode.encode_right(num[4:])
        )



#!/usr/bin/env python
"""Extension for removing the colour red from selected objects"""

import inkex


class RemoveRed(inkex.ColorExtension):
    """Remove red color from selected objects"""

    def modify_color(self, name, color):
        return inkex.Color([0, color.green, color.blue])


if __name__ == "__main__":
    RemoveRed().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 John Beard john.j.beard@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension draws 3d objects from a Wavefront .obj 3D file stored in a local folder
Many settings for appearance, lighting, rotation, etc are available.

                              ^y
                              |
        __--``|               |_--``|     __--
  __--``      |         __--``|     |_--``
 |       z    |        |      |_--``|
 |       <----|--------|-----_0-----|----------------
 |            |        |_--`` |     |
 |      __--``     <-``|      |_--``
 |__--``           x   |__--``|
  IMAGE PLANE           SCENE|
                              |

 Vertices are given as "v" followed by three numbers (x,y,z).
 All files need a vertex list
 v  x.xxx   y.yyy   z.zzz

 Faces are given by a list of vertices
 (vertex 1 is the first in the list above, 2 the second, etc):
 f  1   2   3

 Edges are given by a list of vertices. These will be broken down
 into adjacent pairs automatically.
 l  1   2   3

 Faces are rendered according to the painter's algorithm and perhaps
 back-face culling, if selected. The parameter to sort the faces by
 is user-selectable between max, min and average z-value of the vertices
"""

import os
from math import acos, cos, floor, pi, sin, sqrt

import inkex
from inkex.utils import pairwise
from inkex import Group, Circle
from inkex.paths import Move, Line
from inkex.localization import inkex_gettext as _

try:
    import numpy
except:
    numpy = None


def draw_circle(r, cx, cy, width, fill, name, parent):
    """Draw an SVG circle"""
    circle = parent.add(Circle(cx=str(cx), cy=str(cy), r=str(r)))
    circle.style = {"stroke": "#000000", "stroke-width": str(width), "fill": fill}
    circle.label = name


def draw_line(x1, y1, x2, y2, width, name, parent):
    elem = parent.add(inkex.PathElement())
    elem.style = {
        "stroke": "#000000",
        "stroke-width": str(width),
        "fill": "none",
        "stroke-linecap": "round",
    }
    elem.set("inkscape:label", name)
    elem.path = [Move(x1, y1), Line(x2, y2)]


def draw_poly(pts, face, st, name, parent):
    """Draw polygone"""
    style = {
        "stroke": "#000000",
        "stroke-width": str(st.th),
        "stroke-linejoin": st.linejoin,
        "stroke-opacity": st.s_opac,
        "fill": st.fill,
        "fill-opacity": st.f_opac,
    }
    path = inkex.Path()
    for facet in face:
        if not path:  # for first point
            path.append(Move(pts[facet - 1][0], -pts[facet - 1][1]))
        else:
            path.append(Line(pts[facet - 1][0], -pts[facet - 1][1]))
    path.close()

    poly = parent.add(inkex.PathElement())
    poly.label = name
    poly.style = style
    poly.path = path


def draw_edges(edge_list, pts, st, parent):
    for edge in edge_list:  # for every edge
        pt_1 = pts[edge[0] - 1][0:2]  # the point at the start
        pt_2 = pts[edge[1] - 1][0:2]  # the point at the end
        name = "Edge" + str(edge[0]) + "-" + str(edge[1])
        draw_line(pt_1[0], -pt_1[1], pt_2[0], -pt_2[1], st.th, name, parent)


def draw_faces(faces_data, pts, obj, shading, fill_col, st, parent):
    for face in faces_data:  # for every polygon that has been sorted
        if shading:
            st.fill = get_darkened_colour(
                fill_col, face[1] / pi
            )  # darken proportionally to angle to lighting vector
        else:
            st.fill = get_darkened_colour(fill_col, 1)  # do not darken colour

        face_no = face[3]  # the number of the face to draw
        draw_poly(pts, obj.fce[face_no], st, "Face:" + str(face_no), parent)


def get_darkened_colour(rgb, factor):
    """return a hex triplet of colour, reduced in lightness 0.0-1.0"""
    return (
        "#"
        + "%02X" % floor(factor * rgb[0])
        + "%02X" % floor(factor * rgb[1])
        + "%02X" % floor(factor * rgb[2])
    )  # make the colour string


def make_rotation_log(options):
    """makes a string recording the axes and angles of each rotation, so an object can be repeated"""
    return (
        options.r1_ax
        + str("%.2f" % options.r1_ang)
        + ":"
        + options.r2_ax
        + str("%.2f" % options.r2_ang)
        + ":"
        + options.r3_ax
        + str("%.2f" % options.r3_ang)
        + ":"
        + options.r1_ax
        + str("%.2f" % options.r4_ang)
        + ":"
        + options.r2_ax
        + str("%.2f" % options.r5_ang)
        + ":"
        + options.r3_ax
        + str("%.2f" % options.r6_ang)
    )


def normalise(vector):
    """return the unit vector pointing in the same direction as the argument"""
    length = sqrt(numpy.dot(vector, vector))
    return numpy.array(vector) / length


def get_normal(pts, face):
    """normal vector for the plane passing though the first three elements of face of pts"""
    return numpy.cross(
        (numpy.array(pts[face[0] - 1]) - numpy.array(pts[face[1] - 1])),
        (numpy.array(pts[face[0] - 1]) - numpy.array(pts[face[2] - 1])),
    ).flatten()


def get_unit_normal(pts, face, cw_wound):
    """
    Returns the unit normal for the plane passing through the
    first three points of face, taking account of winding
    """
    # if it is clockwise wound, reverse the vector direction
    winding = -1 if cw_wound else 1
    return winding * normalise(get_normal(pts, face))


def rotate(matrix, rads, axis):
    """choose the correct rotation matrix to use"""
    if axis == "x":
        trans_mat = numpy.array(
            [[1, 0, 0], [0, cos(rads), -sin(rads)], [0, sin(rads), cos(rads)]]
        )
    elif axis == "y":
        trans_mat = numpy.array(
            [[cos(rads), 0, sin(rads)], [0, 1, 0], [-sin(rads), 0, cos(rads)]]
        )
    elif axis == "z":
        trans_mat = numpy.array(
            [[cos(rads), -sin(rads), 0], [sin(rads), cos(rads), 0], [0, 0, 1]]
        )
    return numpy.matmul(trans_mat, matrix)


class Style(object):  # container for style information
    def __init__(self, options):
        self.th = options.th
        self.fill = "#ff0000"
        self.col = "#000000"
        self.r = 2
        self.f_opac = str(options.f_opac / 100.0)
        self.s_opac = str(options.s_opac / 100.0)
        self.linecap = "round"
        self.linejoin = "round"


class WavefrontObj(object):
    """Wavefront based 3d object defined by the vertices and the faces (eg a polyhedron)"""

    name = property(lambda self: self.meta.get("name", None))

    def __init__(self, filename):
        self.meta = {"name": os.path.basename(filename).rsplit(".", 1)[0]}
        self.vtx = []
        self.edg = []
        self.fce = []
        self._parse_file(filename)

    def _parse_file(self, filename):
        if not os.path.isfile(filename):
            raise IOError(_("Can't find wavefront object file {}").format(filename))
        with open(filename, "r") as fhl:
            for line in fhl:
                self._parse_line(line.strip())

    def _parse_line(self, line):
        if line.startswith("#"):
            if ":" in line:
                name, value = line.split(":", 1)
                self.meta[name.lower()] = value
        elif line:
            (kind, line) = line.split(None, 1)
            kind_name = "add_" + kind
            if hasattr(self, kind_name):
                getattr(self, kind_name)(line)

    @staticmethod
    def _parse_numbers(line, typ=str):
        # Ignore any slash options and always pick the first one
        return [typ(v.split("/")[0]) for v in line.split()]

    def add_v(self, line):
        """Add vertex from parsed line"""
        vertex = self._parse_numbers(line, float)
        if len(vertex) == 3:
            self.vtx.append(vertex)

    def add_l(self, line):
        """Add line from parsed line"""
        vtxlist = self._parse_numbers(line, int)
        # we need at least 2 vertices to make an edge
        if len(vtxlist) > 1:
            # we can have more than one vertex per line - get adjacent pairs
            self.edg.extend(pairwise(vtxlist, start=False))

    def add_f(self, line):
        """Add face from parsed line"""
        vtxlist = self._parse_numbers(line, int)
        # we need at least 3 vertices to make an edge
        if len(vtxlist) > 2:
            self.fce.append(vtxlist)

    def get_transformed_pts(self, trans_mat):
        """translate vertex points according to the matrix"""
        transformed_pts = []
        for vtx in self.vtx:
            transformed_pts.append(
                (numpy.matmul(trans_mat, numpy.array(vtx).T)).T.tolist()
            )
        return transformed_pts

    def get_edge_list(self):
        """make an edge vertex list from an existing face vertex list"""
        edge_list = []
        for face in self.fce:
            for j, edge in enumerate(face):
                # Ascending order of certices (for duplicate detection)
                edge_list.append(sorted([edge, face[(j + 1) % len(face)]]))
        return [list(x) for x in sorted(set(tuple(x) for x in edge_list))]


class Poly3D(inkex.GenerateExtension):
    """Generate a polyhedron from a wavefront 3d model file"""

    def add_arguments(self, pars):
        pars.add_argument("--tab", default="common")

        # MODEL FILE SETTINGS
        pars.add_argument("--obj", default="cube")
        pars.add_argument("--spec_file", default="great_rhombicuboct.obj")
        pars.add_argument("--cw_wound", type=inkex.Boolean, default=False)
        pars.add_argument("--type", default="face")
        # VEIW SETTINGS
        pars.add_argument("--r1_ax", default="x")
        pars.add_argument("--r2_ax", default="x")
        pars.add_argument("--r3_ax", default="x")
        pars.add_argument("--r4_ax", default="x")
        pars.add_argument("--r5_ax", default="x")
        pars.add_argument("--r6_ax", default="x")
        pars.add_argument("--r1_ang", type=float, default=0.0)
        pars.add_argument("--r2_ang", type=float, default=0.0)
        pars.add_argument("--r3_ang", type=float, default=0.0)
        pars.add_argument("--r4_ang", type=float, default=0.0)
        pars.add_argument("--r5_ang", type=float, default=0.0)
        pars.add_argument("--r6_ang", type=float, default=0.0)
        pars.add_argument("--scl", type=float, default=100.0)
        # STYLE SETTINGS
        pars.add_argument("--show", type=self.arg_method("gen"))
        pars.add_argument("--shade", type=inkex.Boolean, default=True)
        pars.add_argument("--f_r", type=int, default=255)
        pars.add_argument("--f_g", type=int, default=0)
        pars.add_argument("--f_b", type=int, default=0)
        pars.add_argument("--f_opac", type=int, default=100)
        pars.add_argument("--s_opac", type=int, default=100)
        pars.add_argument("--th", type=float, default=2)
        pars.add_argument("--lv_x", type=float, default=1)
        pars.add_argument("--lv_y", type=float, default=1)
        pars.add_argument("--lv_z", type=float, default=-2)
        pars.add_argument("--back", type=inkex.Boolean, default=False)
        pars.add_argument(
            "--z_sort", type=self.arg_method("z_sort"), default=self.z_sort_min
        )

    def get_filename(self):
        """Get the filename for the spec file"""
        name = ""
        if self.options.obj == "from_file":
            name = self.options.spec_file
        else:
            name = self.options.obj + ".obj"
        moddir = self.ext_path()
        return os.path.join(moddir, "Poly3DObjects", name)

    def generate(self):
        if numpy is None:
            raise inkex.AbortExtension(_("This extension requires the numpy library."))
        so = self.options

        obj = WavefrontObj(self.get_filename())

        scale = self.svg.unittouu("1px")  # convert to document units
        st = Style(so)  # initialise style

        # we will put all the rotations in the object name, so it can be repeated in
        poly = Group.new(obj.name + ":" + make_rotation_log(so))
        (pos_x, pos_y) = self.svg.namedview.center
        poly.transform.add_translate(pos_x, pos_y)
        poly.transform.add_scale(scale)

        # TRANSFORMATION OF THE OBJECT (ROTATION, SCALE, ETC)
        trans_mat = numpy.identity(3, float)  # init. trans matrix as identity matrix
        for i in range(1, 7):  # for each rotation
            axis = getattr(so, "r{}_ax".format(i))
            angle = getattr(so, "r{}_ang".format(i)) * pi / 180
            trans_mat = rotate(trans_mat, angle, axis)
        # scale by linear factor (do this only after the transforms to reduce round-off)
        trans_mat = trans_mat * so.scl

        # the points as projected in the z-axis onto the viewplane
        transformed_pts = obj.get_transformed_pts(trans_mat)
        so.show(obj, st, poly, transformed_pts)
        return poly

    def gen_vtx(self, obj, st, poly, transformed_pts):
        """Generate Vertex"""
        for i, pts in enumerate(transformed_pts):
            draw_circle(st.r, pts[0], pts[1], st.th, "#000000", "Point" + str(i), poly)

    def gen_edg(self, obj, st, poly, transformed_pts):
        """Generate edges"""
        # we already have an edge list
        edge_list = obj.edg
        if obj.fce:
            # we must generate the edge list from the faces
            edge_list = obj.get_edge_list()

        draw_edges(edge_list, transformed_pts, st, poly)

    def gen_fce(self, obj, st, poly, transformed_pts):
        """Generate face"""
        so = self.options
        # colour tuple for the face fill
        fill_col = (so.f_r, so.f_g, so.f_b)
        # unit light vector
        lighting = normalise((so.lv_x, -so.lv_y, so.lv_z))
        # we have a face list
        if obj.fce:
            z_list = []

            for i, face in enumerate(obj.fce):
                # get the normal vector to the face
                norm = get_unit_normal(transformed_pts, face, so.cw_wound)
                # get the angle between the normal and the lighting vector
                angle = acos(numpy.dot(norm, lighting))
                z_sort_param = so.z_sort(transformed_pts, face)

                # include all polygons or just the front-facing ones as needed
                if so.back or norm[2] > 0:
                    # record the maximum z-value of the face and angle to
                    # light, along with the face ID and normal
                    z_list.append((z_sort_param, angle, norm, i))

            z_list.sort(
                key=lambda x: x[0]
            )  # sort by ascending sort parameter of the face
            draw_faces(z_list, transformed_pts, obj, so.shade, fill_col, st, poly)

        else:  # we cannot generate a list of faces from the edges without a lot of computation
            raise inkex.AbortExtension(_("Face data not found."))

    @staticmethod
    def z_sort_max(pts, face):
        """returns the largest z_value of any point in the face"""
        return max([pts[facet - 1][2] for facet in face])

    @staticmethod
    def z_sort_min(pts, face):
        """returns the smallest z_value of any point in the face"""
        return min([pts[facet - 1][2] for facet in face])

    @staticmethod
    def z_sort_cent(pts, face):
        """returns the centroid z_value of any point in the face"""
        return sum([pts[facet - 1][2] for facet in face]) / len(face)


if __name__ == "__main__":
    Poly3D().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 Tavmjong Bah, tavmjong@free.fr
# Copyright (C) 2006 Georg Wiora, xorx@quarkbox.de
# Copyright (C) 2006 Johan Engelen, johan@shouraizou.nl
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Changes:
#  * This program is a modified version of wavy.py by Aaron Spike.
#  * 22-Dec-2006: Wiora : Added axis and isotropic scaling
#  * 21-Jun-2007: Tavmjong: Added polar coordinates
#
from math import cos, pi, sin

import inkex
from inkex import ClipPath, Rectangle
from inkex.utils import math_eval

from inkex.localization import inkex_gettext as _


def drawfunction(
    xstart,
    xend,
    ybottom,
    ytop,
    samples,
    width,
    height,
    left,
    bottom,
    fx="sin(x)",
    fpx="cos(x)",
    fponum=True,
    times2pi=False,
    polar=False,
    isoscale=True,
    drawaxis=True,
    endpts=False,
):
    if times2pi:
        xstart = 2 * pi * xstart
        xend = 2 * pi * xend

    # coords and scales based on the source rect
    if xstart == xend:
        inkex.errormsg(
            _(
                "x-interval cannot be zero. Please modify 'Start X value' "
                "or 'End X value'"
            )
        )
        return []
    scalex = width / (xend - xstart)
    xoff = left
    coordx = lambda x: (x - xstart) * scalex + xoff  # convert x-value to coordinate
    if polar:  # Set scale so that left side of rectangle is -1, right side is +1.
        # (We can't use xscale for both range and scale.)
        centerx = left + width / 2.0
        polar_scalex = width / 2.0
        coordx = lambda x: x * polar_scalex + centerx  # convert x-value to coordinate

    if ytop == ybottom:
        inkex.errormsg(
            _(
                "y-interval cannot be zero. Please modify 'Y value of rectangle's top' "
                "or 'Y value of rectangle's bottom'"
            )
        )
        return []
    scaley = height / (ytop - ybottom)
    yoff = bottom
    coordy = lambda y: (ybottom - y) * scaley + yoff  # convert y-value to coordinate

    # Check for isotropic scaling and use smaller of the two scales, correct ranges
    if isoscale and not polar:
        if scaley < scalex:
            # compute zero location
            xzero = coordx(0)
            # set scale
            scalex = scaley
            # correct x-offset
            xstart = (left - xzero) / scalex
            xend = (left + width - xzero) / scalex
        else:
            # compute zero location
            yzero = coordy(0)
            # set scale
            scaley = scalex
            # correct x-offset
            ybottom = (yzero - bottom) / scaley
            ytop = (bottom + height - yzero) / scaley

    f = math_eval(fx)
    fp = math_eval(fpx)
    if f is None or (fp is None and not (fponum)):
        raise inkex.AbortExtension(_("Invalid function specification"))

    # step is the distance between nodes on x
    step = (xend - xstart) / (samples - 1)
    third = step / 3.0
    ds = step * 0.001  # Step used in calculating derivatives

    a = []  # path array
    # add axis
    if drawaxis:
        # check for visibility of x-axis
        if ybottom <= 0 <= ytop:
            # xaxis
            a.append(["M", [left, coordy(0)]])
            a.append(["l", [width, 0]])
        # check for visibility of y-axis
        if xstart <= 0 <= xend:
            # xaxis
            a.append(["M", [coordx(0), bottom]])
            a.append(["l", [0, -height]])

    # initialize function and derivative for 0;
    # they are carried over from one iteration to the next, to avoid extra function calculations.
    x0 = xstart
    y0 = f(xstart)
    if polar:
        xp0 = y0 * cos(x0)
        yp0 = y0 * sin(x0)
        x0 = xp0
        y0 = yp0
    if (
        fponum or polar
    ):  # numerical derivative, using 0.001*step as the small differential
        x1 = xstart + ds  # Second point AFTER first point (Good for first point)
        y1 = f(x1)
        if polar:
            xp1 = y1 * cos(x1)
            yp1 = y1 * sin(x1)
            x1 = xp1
            y1 = yp1
        dx0 = (x1 - x0) / ds
        dy0 = (y1 - y0) / ds
    else:  # derivative given by the user
        dx0 = 1  # Only works for rectangular coordinates
        dy0 = fp(xstart)

    # Start curve
    if endpts:
        a.append(["M", [left, coordy(0)]])
        a.append(["L", [coordx(x0), coordy(y0)]])
    else:
        a.append(["M", [coordx(x0), coordy(y0)]])  # initial moveto

    for i in range(int(samples - 1)):
        x1 = (i + 1) * step + xstart
        x2 = x1 - ds  # Second point BEFORE first point (Good for last point)
        y1 = f(x1)
        y2 = f(x2)
        if polar:
            xp1 = y1 * cos(x1)
            yp1 = y1 * sin(x1)
            xp2 = y2 * cos(x2)
            yp2 = y2 * sin(x2)
            x1 = xp1
            y1 = yp1
            x2 = xp2
            y2 = yp2
        if fponum or polar:  # numerical derivative
            dx1 = (x1 - x2) / ds
            dy1 = (y1 - y2) / ds
        else:  # derivative given by the user
            dx1 = 1  # Only works for rectangular coordinates
            dy1 = fp(x1)
        # create curve
        a.append(
            [
                "C",
                [
                    coordx(x0 + (dx0 * third)),
                    coordy(y0 + (dy0 * third)),
                    coordx(x1 - (dx1 * third)),
                    coordy(y1 - (dy1 * third)),
                    coordx(x1),
                    coordy(y1),
                ],
            ]
        )
        x0 = x1  # Next segment's start is this segments end
        y0 = y1
        dx0 = dx1  # Assume the function is smooth everywhere, so carry over the derivative too
        dy0 = dy1
    if endpts:
        a.append(["L", [left + width, coordy(0)]])
    return a


class FuncPlot(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--xstart", type=float, default=0.0, help="Start x-value")
        pars.add_argument("--xend", type=float, default=1.0, help="End x-value")
        pars.add_argument(
            "--times2pi", type=inkex.Boolean, default=False, help="* x-range by 2*pi"
        )
        pars.add_argument(
            "--polar", type=inkex.Boolean, default=False, help="Use polar coords"
        )
        pars.add_argument(
            "--ybottom", type=float, default=0.0, help="y-value of rect's bottom"
        )
        pars.add_argument(
            "--ytop", type=float, default=1.0, help="y-value of rectangle's top"
        )
        pars.add_argument("--samples", type=int, default=8, help="Samples")
        pars.add_argument("--fofx", default="sin(x)", help="f(x) for plotting")
        pars.add_argument(
            "--fponum", type=inkex.Boolean, default=True, help="Numerical 1st deriv"
        )
        pars.add_argument("--fpofx", default="cos(x)", help="f'(x) for plotting")
        pars.add_argument(
            "--clip", type=inkex.Boolean, default=False, help="Clip with source rect"
        )
        pars.add_argument(
            "--remove", type=inkex.Boolean, default=True, help="Remove source rect"
        )
        pars.add_argument(
            "--isoscale", type=inkex.Boolean, default=True, help="Isotropic scaling"
        )
        pars.add_argument(
            "--drawaxis", type=inkex.Boolean, default=False, help="Draw axis"
        )
        pars.add_argument(
            "--endpts", type=inkex.Boolean, default=False, help="Add end points"
        )

    def effect(self):
        newpath = None
        for node in self.svg.selection.values():
            if isinstance(node, Rectangle):
                # create new path with basic dimensions of selected rectangle
                newpath = inkex.PathElement()
                x = float(node.get("x"))
                y = float(node.get("y"))
                w = float(node.get("width"))
                h = float(node.get("height"))

                # copy attributes of rect
                newpath.style = node.style
                newpath.transform = node.transform

                # top and bottom were exchanged
                newpath.path = drawfunction(
                    self.options.xstart,
                    self.options.xend,
                    self.options.ybottom,
                    self.options.ytop,
                    self.options.samples,
                    w,
                    h,
                    x,
                    y + h,
                    self.options.fofx,
                    self.options.fpofx,
                    self.options.fponum,
                    self.options.times2pi,
                    self.options.polar,
                    self.options.isoscale,
                    self.options.drawaxis,
                    self.options.endpts,
                )
                newpath.set("title", self.options.fofx)

                # add path into SVG structure
                node.getparent().append(newpath)
                # option whether to clip the path with rect or not.
                if self.options.clip:
                    clip = self.svg.defs.add(ClipPath())
                    clip.set_random_id()
                    clip.append(node.copy())
                    newpath.set("clip-path", clip.get_id(as_url=2))
                # option whether to remove the rectangle or not.
                if self.options.remove:
                    node.getparent().remove(node)
        if newpath is None:
            raise inkex.AbortExtension(_("Please select a rectangle"))


if __name__ == "__main__":
    FuncPlot().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2011 Nicolas Dufour (jazzynico)
# Direction code from the Restack extension, by Rob Antonishen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""
Extract text and print it to the error console.
"""

from lxml.etree import tostring

import inkex
from inkex import TextElement, FlowRoot
from inkex.utils import KeyDict

# Old settings, supported because users click 'ok' without looking.
XAN = KeyDict({"l": "left", "r": "right", "m": "center_x"})
YAN = KeyDict({"t": "top", "b": "bottom", "m": "center_y"})


class Extract(inkex.EffectExtension):
    """Extract text and print out"""

    select_all = (TextElement, FlowRoot)

    def add_arguments(self, pars):
        pars.add_argument(
            "-d", "--direction", default="lr", help="direction to extract text"
        )
        pars.add_argument(
            "-x", "--xanchor", default="left", help="horiz point to compare"
        )
        pars.add_argument(
            "-y", "--yanchor", default="top", help="vertical point to compare"
        )

    def effect(self):
        # move them to the top of the object stack in this order.
        for node in sorted(
            self.svg.selection.get(TextElement, FlowRoot), key=self._sort
        ):
            self.recurse(node)

    def _sort(self, node):
        return node.bounding_box().get_anchor(
            self.options.xanchor, self.options.yanchor, self.options.direction
        )

    def recurse(self, node):
        """Go through each node and recusively self call for all children"""
        if node.text is not None or node.tail is not None:
            for child in node:
                if child.get("sodipodi:role"):
                    child.tail = "\n"
            inkex.errormsg(tostring(node, encoding="unicode", method="text").strip())
        else:
            for child in node:
                self.recurse(child)


if __name__ == "__main__":
    Extract().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Tavmjong Bah
#               2019 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Generic template functionality controlled by the INX file.
"""

import inkex


class DvdCover(inkex.TemplateExtension):
    """Create an empty DVD Cover (in mm)"""

    multi_inx = True

    def add_arguments(self, pars):
        pars.add_argument(
            "-s", "--spine", type=float, default=14.0, help="Dvd spine width (mm)"
        )
        pars.add_argument(
            "-b",
            "--bleed",
            type=float,
            default=3.0,
            help="Bleed (extra area around image",
        )

    def get_size(self):
        # Dimensions in mm
        width = 259.0  # Before adding spine width or bleed
        height = 183.0  # Before adding bleed
        bleed = self.options.bleed
        spine = self.options.spine

        return (width + spine + bleed * 2.0, "mm", height + bleed * 2.0, "mm")

    def set_namedview(self, width, height, unit):
        super(DvdCover, self).set_namedview(width, height, unit)
        (width, _, height, unit) = self.get_size()
        bleed, spine = self.options.bleed, self.options.spine
        self.svg.namedview.new_guide(bleed, True, "bottom")
        self.svg.namedview.new_guide(height - bleed, True, "top")
        self.svg.namedview.new_guide(bleed, False, "left edge")
        self.svg.namedview.new_guide((width - spine) / 2.0, False, "left spline")
        self.svg.namedview.new_guide((width + spine) / 2.0, False, "right spline")
        self.svg.namedview.new_guide(width - bleed, False, "top")


if __name__ == "__main__":
    DvdCover().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""
JessyInk Export to zipfile multiple layers.
"""

import zipfile

import inkex
from inkex.base import TempDirMixin
from inkex.command import take_snapshot

from jessyink_install import JessyInkMixin


class Export(JessyInkMixin, TempDirMixin, inkex.OutputExtension):
    """
    JessyInkExport Output Extension saves to a zipfile each of the layers.
    """

    dir_prefix = "jessyInk-"

    def add_arguments(self, pars):
        pars.add_argument("--tab", type=str, dest="what")
        pars.add_argument("--type", type=str, dest="type", default="pdf")
        pars.add_argument("--resolution", type=int, default=92)

    def save(self, stream):
        self.is_installed()

        with zipfile.ZipFile(stream, "w", compression=zipfile.ZIP_STORED) as output:

            # Find layers.
            layers = self.svg.xpath("//svg:g[@inkscape:groupmode='layer']")

            if len(layers) < 1:
                inkex.errormsg("No layers found.")
                return

            for node in layers:
                # Make all layers invisible
                node.style["display"] = "none"

            for node in layers:
                # Show only one layer at a time.
                node.style.update("display:inherit;opacity:1")

                name = node.get("inkscape:label")
                newname = "{}.{}".format(name, self.options.type)
                filename = take_snapshot(
                    self.document,
                    dirname=self.tempdir,
                    name=name,
                    ext=self.options.type,
                    dpi=self.options.resolution,
                )
                output.write(filename, newname)

                node.style["display"] = "none"


if __name__ == "__main__":
    Export().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2011 Nikita Kitaev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
An Inkscape extension for exporting Synfig files (.sif)
"""
import math
import uuid
from copy import deepcopy

from lxml import etree

import inkex
from inkex import (
    Group,
    Layer,
    Anchor,
    Switch,
    PathElement,
    Metadata,
    NamedView,
    Gradient,
    SvgDocumentElement,
    Path,
    Transform,
)

import synfig_fileformat as sif
from synfig_prepare import MalformedSVGError, SynfigPrep, get_dimension


# ##### Utility Classes ####################################
class UnsupportedException(Exception):
    """When part of an element is not supported, this exception is raised to invalidate the whole element"""

    pass


class SynfigDocument(object):
    """A synfig document, with commands for adding layers and layer parameters"""

    def __init__(self, width=1024, height=768, name="Synfig Animation 1"):
        self.root_canvas = etree.fromstring(
            """
    <canvas
        version="0.5"
        width="{:f}"
        height="{:f}"
        xres="2834.645752"
        yres="2834.645752"
        view-box="0 0 0 0"
        >
      <name>{}</name>
    </canvas>
    """.format(
                width, height, name
            )
        )

        self._update_viewbox()

        self.gradients = {}
        self.filters = {}

    # ## Properties

    def get_root_canvas(self):
        return self.root_canvas

    def get_root_tree(self):
        return self.root_canvas.getroottree()

    def _update_viewbox(self):
        """Update the viewbox to match document width and height"""
        attr_viewbox = "{:f} {:f} {:f} {:f}".format(
            -self.width / 2.0 / sif.kux,
            self.height / 2.0 / sif.kux,
            self.width / 2.0 / sif.kux,
            -self.height / 2.0 / sif.kux,
        )
        self.root_canvas.set("view-box", attr_viewbox)

    def get_width(self):
        return float(self.root_canvas.get("width", "0"))

    def set_width(self, value):
        self.root_canvas.set("width", str(value))
        self._update_viewbox()

    def get_height(self):
        return float(self.root_canvas.get("height", "0"))

    def set_height(self, value):
        self.root_canvas.set("height", str(value))
        self._update_viewbox()

    def get_name(self):
        return self.root_canvas.get("name", "")

    def set_name(self, value):
        self.root_canvas.set("name", value)
        self._update_viewbox()

    width = property(get_width, set_width)
    height = property(get_height, set_height)
    name = property(get_name, set_name)

    # ## Public utility functions

    def new_guid(self):
        """Generate a new GUID"""
        return uuid.uuid4().hex

    # ## Coordinate system conversions

    def distance_svg2sif(self, distance):
        """Convert distance from SVG to Synfig units"""
        return distance / sif.kux

    def distance_sif2svg(self, distance):
        """Convert distance from Synfig to SVG units"""
        return distance * sif.kux

    def coor_svg2sif(self, vector):
        """Convert SVG coordinate [x, y] to Synfig units"""
        x = vector[0]
        y = self.height - vector[1]

        x -= self.width / 2.0
        y -= self.height / 2.0
        x /= sif.kux
        y /= sif.kux

        return [x, y]

    def coor_sif2svg(self, vector):
        """Convert Synfig coordinate [x, y] to SVG units"""
        x = vector[0] * sif.kux + self.width / 2.0
        y = vector[1] * sif.kux + self.height / 2.0

        y = self.height - y

        assert (
            self.coor_svg2sif([x, y]) == vector
        ), "sif to svg coordinate conversion error"

        return [x, y]

    def list_coor_svg2sif(self, l):
        """Scan a list for coordinate pairs and convert them to Synfig units"""
        # If list has two numerical elements,
        # treat it as a coordinate pair
        if type(l) == list and len(l) == 2:
            if type(l[0]) == int or type(l[0]) == float:
                if type(l[1]) == int or type(l[1]) == float:
                    l_sif = self.coor_svg2sif(l)
                    l[0] = l_sif[0]
                    l[1] = l_sif[1]
                    return

        # Otherwise recursively iterate over the list
        for x in l:
            if type(x) == list:
                self.list_coor_svg2sif(x)

    def list_coor_sif2svg(self, l):
        """Scan a list for coordinate pairs and convert them to SVG units"""
        # If list has two numerical elements,
        # treat it as a coordinate pair
        if type(l) == list and len(l) == 2:
            if type(l[0]) == int or type(l[0]) == float:
                if type(l[1]) == int or type(l[1]) == float:
                    l_sif = self.coor_sif2svg(l)
                    l[0] = l_sif[0]
                    l[1] = l_sif[1]
                    return

        # Otherwise recursively iterate over the list
        for x in l:
            if type(x) == list:
                self.list_coor_sif2svg(x)

    def bline_coor_svg2sif(self, b):
        """Convert a BLine from SVG to Synfig coordinate units"""
        self.list_coor_svg2sif(b["points"])

    def bline_coor_sif2svg(self, b):
        """Convert a BLine from Synfig to SVG coordinate units"""
        self.list_coor_sif2svg(b["points"])

    # ## XML Builders -- private
    # ##  used to create XML elements in the Synfig document

    def build_layer(self, layer_type, desc, canvas=None, active=True, version="auto"):
        """Build an empty layer"""
        if canvas is None:
            layer = self.root_canvas.makeelement("layer")
        else:
            layer = etree.SubElement(canvas, "layer")

        layer.set("type", layer_type)
        layer.set("desc", desc)
        if active:
            layer.set("active", "true")
        else:
            layer.set("active", "false")

        if version == "auto":
            version = sif.defaultLayerVersion(layer_type)

        if type(version) == float:
            version = str(version)

        layer.set("version", version)

        return layer

    def _calc_radius(self, p1x, p1y, p2x, p2y):
        """Calculate radius of a tangent given two points"""
        # Synfig tangents are scaled by a factor of 3
        return sif.tangent_scale * math.sqrt((p2x - p1x) ** 2 + (p2y - p1y) ** 2)

    def _calc_angle(self, p1x, p1y, p2x, p2y):
        """Calculate angle (in radians) of a tangent given two points"""
        dx = p2x - p1x
        dy = p2y - p1y
        if dx > 0 and dy > 0:
            ag = math.pi + math.atan(dy / dx)
        elif dx > 0 > dy:
            ag = math.pi + math.atan(dy / dx)
        elif dx < 0 and dy < 0:
            ag = math.atan(dy / dx)
        elif dx < 0 < dy:
            ag = 2 * math.pi + math.atan(dy / dx)
        elif dx == 0 and dy > 0:
            ag = -1 * math.pi / 2
        elif dx == 0 and dy < 0:
            ag = math.pi / 2
        elif dx == 0 and dy == 0:
            ag = 0
        elif dx < 0 and dy == 0:
            ag = 0
        elif dx > 0 and dy == 0:
            ag = math.pi

        return (ag * 180) / math.pi

    def build_param(self, layer, name, value, param_type="auto", guid=None):
        """Add a parameter node to a layer"""
        if layer is None:
            param = self.root_canvas.makeelement("param")
        else:
            param = etree.SubElement(layer, "param")
        param.set("name", name)

        # Automatically detect param_type
        if param_type == "auto":
            if layer is not None:
                layer_type = layer.get("type")
                param_type = sif.paramType(layer_type, name)
            else:
                param_type = sif.paramType(None, name, value)

        if param_type == "real":
            el = etree.SubElement(param, "real")
            el.set("value", str(float(value)))
        elif param_type == "integer":
            el = etree.SubElement(param, "integer")
            el.set("value", str(int(value)))
        elif param_type == "vector":
            el = etree.SubElement(param, "vector")
            x = etree.SubElement(el, "x")
            x.text = str(float(value[0]))
            y = etree.SubElement(el, "y")
            y.text = str(float(value[1]))
        elif param_type == "color":
            el = etree.SubElement(param, "color")
            r = etree.SubElement(el, "r")
            r.text = str(float(value[0]))
            g = etree.SubElement(el, "g")
            g.text = str(float(value[1]))
            b = etree.SubElement(el, "b")
            b.text = str(float(value[2]))
            a = etree.SubElement(el, "a")
            a.text = str(float(value[3])) if len(value) > 3 else "1.0"
        elif param_type == "gradient":
            el = etree.SubElement(param, "gradient")
            # Value is a dictionary of color stops
            #  see get_gradient()
            for pos in value.keys():
                color = etree.SubElement(el, "color")
                color.set("pos", str(float(pos)))

                c = value[pos]

                r = etree.SubElement(color, "r")
                r.text = str(float(c[0]))
                g = etree.SubElement(color, "g")
                g.text = str(float(c[1]))
                b = etree.SubElement(color, "b")
                b.text = str(float(c[2]))
                a = etree.SubElement(color, "a")
                a.text = str(float(c[3])) if len(c) > 3 else "1.0"
        elif param_type == "bool":
            el = etree.SubElement(param, "bool")
            if value:
                el.set("value", "true")
            else:
                el.set("value", "false")
        elif param_type == "time":
            el = etree.SubElement(param, "time")
            if type(value) == int:
                el.set("value", "{:d}s".format(value))
            elif type(value) == float:
                el.set("value", "{:f}s".format(value))
            elif type(value) == str:
                el.set("value", value)
        elif param_type == "bline":
            el = etree.SubElement(param, "bline")
            el.set("type", "bline_point")

            # value is a bline (dictionary type), see path_to_bline_list
            if value["loop"]:
                el.set("loop", "true")
            else:
                el.set("loop", "false")

            for vertex in value["points"]:
                x = float(vertex[1][0])
                y = float(vertex[1][1])

                tg1x = float(vertex[0][0])
                tg1y = float(vertex[0][1])

                tg2x = float(vertex[2][0])
                tg2y = float(vertex[2][1])

                tg1_radius = self._calc_radius(x, y, tg1x, tg1y)
                tg1_angle = self._calc_angle(x, y, tg1x, tg1y)

                tg2_radius = self._calc_radius(x, y, tg2x, tg2y)
                tg2_angle = self._calc_angle(x, y, tg2x, tg2y) - 180.0

                if vertex[3]:
                    split = "true"
                else:
                    split = "false"

                entry = etree.SubElement(el, "entry")
                composite = etree.SubElement(entry, "composite")
                composite.set("type", "bline_point")

                point = etree.SubElement(composite, "point")
                vector = etree.SubElement(point, "vector")
                etree.SubElement(vector, "x").text = str(x)
                etree.SubElement(vector, "y").text = str(y)

                width = etree.SubElement(composite, "width")
                etree.SubElement(width, "real").set("value", "1.0")

                origin = etree.SubElement(composite, "origin")
                etree.SubElement(origin, "real").set("value", "0.5")

                split_el = etree.SubElement(composite, "split")
                etree.SubElement(split_el, "bool").set("value", split)

                t1 = etree.SubElement(composite, "t1")
                t2 = etree.SubElement(composite, "t2")

                t1_rc = etree.SubElement(t1, "radial_composite")
                t1_rc.set("type", "vector")

                t2_rc = etree.SubElement(t2, "radial_composite")
                t2_rc.set("type", "vector")

                t1_r = etree.SubElement(t1_rc, "radius")
                t2_r = etree.SubElement(t2_rc, "radius")
                t1_radius = etree.SubElement(t1_r, "real")
                t2_radius = etree.SubElement(t2_r, "real")
                t1_radius.set("value", str(tg1_radius))
                t2_radius.set("value", str(tg2_radius))

                t1_t = etree.SubElement(t1_rc, "theta")
                t2_t = etree.SubElement(t2_rc, "theta")
                t1_angle = etree.SubElement(t1_t, "angle")
                t2_angle = etree.SubElement(t2_t, "angle")
                t1_angle.set("value", str(tg1_angle))
                t2_angle.set("value", str(tg2_angle))
        elif param_type == "canvas":
            el = etree.SubElement(param, "canvas")
            el.set("xres", "10.0")
            el.set("yres", "10.0")

            # "value" is a list of layers
            if value is not None:
                for layer in value:
                    el.append(layer)
        else:
            raise AssertionError("Unsupported param type {}".format(param_type))

        if guid:
            el.set("guid", guid)
        else:
            el.set("guid", self.new_guid())

        return param

    # ## Public layer API
    # ##  Should be used by outside functions to create layers and set layer parameters

    def create_layer(
        self,
        layer_type,
        desc,
        params={},
        guids={},
        canvas=None,
        active=True,
        version="auto",
    ):
        """Create a new layer

        Keyword arguments:
        layer_type -- layer type string used internally by Synfig
        desc -- layer description
        params -- a dictionary of parameter names and their values
        guids -- a dictionary of parameter types and their guids (optional)
        active -- set to False to create a hidden layer
        """
        layer = self.build_layer(layer_type, desc, canvas, active, version)
        default_layer_params = sif.defaultLayerParams(layer_type)

        for param_name in default_layer_params.keys():
            param_type = default_layer_params[param_name][0]
            if param_name in params.keys():
                param_value = params[param_name]
            else:
                param_value = default_layer_params[param_name][1]

            if param_name in guids.keys():
                param_guid = guids[param_name]
            else:
                param_guid = None

            if param_value is not None:
                self.build_param(
                    layer, param_name, param_value, param_type, guid=param_guid
                )

        return layer

    def set_param(
        self, layer, name, value, param_type="auto", guid=None, modify_linked=False
    ):
        """Set a layer parameter

        Keyword arguments:
        layer -- the layer to set the parameter for
        name -- parameter name
        value -- parameter value
        param_type -- parameter type (default "auto")
        guid -- guid of the parameter value
        """
        if modify_linked:
            raise AssertionError("Modifying linked parameters is not supported")

        layer_type = layer.get("type")
        assert layer_type, "Layer does not have a type"

        if param_type == "auto":
            param_type = sif.paramType(layer_type, name)

        # Remove existing parameters with this name
        existing = []
        for param in layer.iterchildren():
            if param.get("name") == name:
                existing.append(param)

        if len(existing) == 0:
            self.build_param(layer, name, value, param_type, guid)
        elif len(existing) > 1:
            raise AssertionError("Found multiple parameters with the same name")
        else:
            new_param = self.build_param(None, name, value, param_type, guid)
            layer.replace(existing[0], new_param)

    def set_params(self, layer, params={}, guids={}, modify_linked=False):
        """Set layer parameters

        Keyword arguments:
        layer -- the layer to set the parameter for
        params -- a dictionary of parameter names and their values
        guids -- a dictionary of parameter types and their guids (optional)
        """
        for param_name in params.keys():
            if param_name in guids.keys():
                self.set_param(
                    layer,
                    param_name,
                    params[param_name],
                    guid=guids[param_name],
                    modify_linked=modify_linked,
                )
            else:
                self.set_param(
                    layer, param_name, params[param_name], modify_linked=modify_linked
                )

    def get_param(self, layer, name, param_type="auto"):
        """Get the value of a layer parameter

        Keyword arguments:
        layer -- the layer to get the parameter from
        name -- param name
        param_type -- parameter type (default "auto")

        NOT FULLY IMPLEMENTED
        """
        layer_type = layer.get("type")
        assert layer_type, "Layer does not have a type"

        if param_type == "auto":
            param_type = sif.paramType(layer_type, name)

        for param in layer.iterchildren():
            if param.get("name") == name:
                if param_type == "real":
                    return float(param[0].get("value", "0"))
                elif param_type == "integer":
                    return int(param[0].get("integer", "0"))
                else:
                    raise Exception(
                        "Getting this type of parameter not yet implemented"
                    )

    # ## Global defs, and related

    # SVG Filters
    def add_filter(self, filter_id, f):
        """Register a filter"""
        self.filters[filter_id] = f

    # SVG Gradients
    def add_linear_gradient(
        self,
        gradient_id,
        p1,
        p2,
        mtx=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
        stops=[],
        link="",
        spread_method="pad",
    ):
        """Register a linear gradient definition"""
        gradient = {
            "type": "linear",
            "p1": p1,
            "p2": p2,
            "mtx": mtx,
            "spreadMethod": spread_method,
        }
        if stops:
            gradient["stops"] = stops
            gradient["stops_guid"] = self.new_guid()
        elif link != "":
            gradient["link"] = link
        else:
            raise MalformedSVGError("Gradient has neither stops nor link")
        self.gradients[gradient_id] = gradient

    def add_radial_gradient(
        self,
        gradient_id,
        center,
        radius,
        focus,
        mtx=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
        stops=[],
        link="",
        spread_method="pad",
    ):
        """Register a radial gradient definition"""
        gradient = {
            "type": "radial",
            "center": center,
            "radius": radius,
            "focus": focus,
            "mtx": mtx,
            "spreadMethod": spread_method,
        }
        if stops:
            gradient["stops"] = stops
            gradient["stops_guid"] = self.new_guid()
        elif link != "":
            gradient["link"] = link
        else:
            raise MalformedSVGError("Gradient has neither stops nor link")
        self.gradients[gradient_id] = gradient

    def get_gradient(self, gradient_id):
        """
        Return a gradient with a given id

        Linear gradient format:
        {
        "type"      : "linear",
        "p1"        : [x, y],
        "p2"        : [x, y],
        "mtx"       : mtx,
        "stops"     : color stops,
        "stops_guid": color stops guid,
        "spreadMethod": "pad", "reflect", or "repeat"
        }

        Radial gradient format:
        {
        "type"      : "radial",
        "center"    : [x, y],
        "radius"    : r,
        "focus"     : [x, y],
        "mtx"       : mtx,
        "stops"     : color stops,
        "stops_guid": color stops guid,
        "spreadMethod": "pad", "reflect", or "repeat"
        }

        Color stops format
        {
        0.0         : color ([r,g,b,a] or [r,g,b]) at start,
        [a number]  : color at that position,
        1.0         : color at end
        }
        """

        if gradient_id not in self.gradients.keys():
            return None

        gradient = self.gradients[gradient_id]

        # If the gradient has no link, we are done
        if "link" not in gradient.keys() or gradient["link"] == "":
            return gradient

        # If the gradient does have a link, find the color stops recursively
        if gradient["link"] not in self.gradients.keys():
            raise MalformedSVGError("Linked gradient ID not found")

        linked_gradient = self.get_gradient(gradient["link"])
        gradient["stops"] = linked_gradient["stops"]
        gradient["stops_guid"] = linked_gradient["stops_guid"]
        del gradient["link"]

        # Update the gradient in our listing
        # (so recursive lookup only happens once)
        self.gradients[gradient_id] = gradient

        return gradient

    def gradient_to_params(self, gradient):
        """Transform gradient to a list of parameters to pass to a Synfig layer"""
        # Create a copy of the gradient
        g = gradient.copy()

        # Set synfig-only attribs
        if g["spreadMethod"] == "repeat":
            g["loop"] = True
        elif g["spreadMethod"] == "reflect":
            g["loop"] = True
            # Reflect the gradient
            # Original: 0.0 [A . B . C] 1.0
            # New:      0.0 [A . B . C . B . A] 1.0
            #           (with gradient size doubled)
            new_stops = {}

            # reflect the stops
            for pos in g["stops"]:
                val = g["stops"][pos]
                if pos == 1.0:
                    new_stops[pos / 2.0] = val
                else:
                    new_stops[pos / 2.0] = val
                    new_stops[1 - pos / 2.0] = val
            g["stops"] = new_stops

            # double the gradient size
            if g["type"] == "linear":
                g["p2"] = [
                    g["p1"][0] + 2.0 * (g["p2"][0] - g["p1"][0]),
                    g["p1"][1] + 2.0 * (g["p2"][1] - g["p1"][1]),
                ]
            if g["type"] == "radial":
                g["radius"] *= 2.0

        # Rename "stops" to "gradient"
        g["gradient"] = g["stops"]

        # Convert coordinates
        if g["type"] == "linear":
            g["p1"] = self.coor_svg2sif(g["p1"])
            g["p2"] = self.coor_svg2sif(g["p2"])

        if g["type"] == "radial":
            g["center"] = self.coor_svg2sif(g["center"])
            g["radius"] = self.distance_svg2sif(g["radius"])

        # Delete extra attribs
        removed_attribs = [
            "type",
            "stops",
            "stops_guid",
            "mtx",
            "focus",
            "spreadMethod",
        ]
        for x in removed_attribs:
            if x in g.keys():
                del g[x]
        return g

    # ## Public operations API
    # Operations act on a series of layers, and (optionally) on a series of named parameters
    # The "is_end" attribute should be set to true when the layers are at the end of a canvas
    # (i.e. when adding transform layers on top of them does not require encapsulation)

    def op_blur(self, layers, x, y, name="Blur", is_end=False):
        """Gaussian blur the given layers by the given x and y amounts

        Keyword arguments:
        layers -- list of layers
        x -- x-amount of blur
        y -- x-amount of blur
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        blur = self.create_layer(
            "blur",
            name,
            params={"blend_method": sif.blend_methods["straight"], "size": [x, y]},
        )

        if is_end:
            return layers + [blur]
        else:
            return self.op_encapsulate(layers + [blur])

    def op_color(self, layers, overlay, is_end=False):
        """Apply a color overlay to the given layers

        Should be used to apply a gradient or pattern to a shape

        Keyword arguments:
        layers -- list of layers
        overlay -- color layer to apply
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        if not layers:
            return layers
        if overlay is None:
            return layers

        overlay_enc = self.op_encapsulate([overlay])
        self.set_param(
            overlay_enc[0], "blend_method", sif.blend_methods["straight onto"]
        )
        ret = layers + overlay_enc

        if is_end:
            return ret
        else:
            return self.op_encapsulate(ret)

    def op_encapsulate(self, layers, name="Inline Canvas", is_end=False):
        """Encapsulate the given layers

        Keyword arguments:
        layers -- list of layers
        name -- Name of the PasteCanvas layer that is created
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of one layer
        """

        if not layers:
            return layers

        layer = self.create_layer("PasteCanvas", name, params={"canvas": layers})
        return [layer]

    def op_fade(self, layers, opacity, is_end=False):
        """Increase the opacity of the given layers by a certain amount

        Keyword arguments:
        layers -- list of layers
        opacity -- the opacity to apply (float between 0.0 to 1.0)
        name -- name of the Transform layer that is added
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        # If there is blending involved, first encapsulate the layers
        for layer in layers:
            if self.get_param(layer, "blend_method") != sif.blend_methods["composite"]:
                return self.op_fade(self.op_encapsulate(layers), opacity, is_end)

        # Otherwise, set their amount
        for layer in layers:
            amount = self.get_param(layer, "amount")
            self.set_param(layer, "amount", amount * opacity)

        return layers

    def op_filter(self, layers, filter_id, is_end=False):
        """Apply a filter to the given layers

        Keyword arguments:
        layers -- list of layers
        filter_id -- id of the filter
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        if filter_id not in self.filters.keys():
            raise MalformedSVGError("Filter {} not found".format(filter_id))

        try:
            ret = self.filters[filter_id](self, layers, is_end)
            assert type(ret) == list
            return ret
        except UnsupportedException:
            # If the filter is not supported, ignore it.
            return layers

    def op_set_blend(self, layers, blend_method, is_end=False):
        """Set the blend method of the given group of layers

        If more than one layer is supplied, they will be encapsulated.

        Keyword arguments:
        layers -- list of layers
        blend_method -- blend method to give the layers
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        if not layers:
            return layers
        if blend_method == "composite":
            return layers

        layer = layers[0]
        if len(layers) > 1 or self.get_param(layers[0], "amount") != 1.0:
            layer = self.op_encapsulate(layers)[0]

        layer = deepcopy(layer)

        self.set_param(layer, "blend_method", sif.blend_methods[blend_method])

        return [layer]

    def op_transform(self, layers, mtx, name="Transform", is_end=False):
        """Apply a matrix transformation to the given layers

        Keyword arguments:
        layers -- list of layers
        mtx -- transformation matrix
        name -- name of the Transform layer that is added
        is_end -- set to True if layers are at the end of a canvas

        Returns: list of layers
        """
        if not layers:
            return layers
        if mtx is None or mtx == [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]:
            return layers

        src_tl = [100, 100]
        src_br = [200, 200]

        dest_tl = [100, 100]
        dest_tr = [200, 100]
        dest_br = [200, 200]
        dest_bl = [100, 200]

        dest_tl = Transform(mtx).apply_to_point(dest_tl)
        dest_tr = Transform(mtx).apply_to_point(dest_tr)
        dest_br = Transform(mtx).apply_to_point(dest_br)
        dest_bl = Transform(mtx).apply_to_point(dest_bl)

        warp = self.create_layer(
            "warp",
            name,
            params={
                "src_tl": self.coor_svg2sif(src_tl),
                "src_br": self.coor_svg2sif(src_br),
                "dest_tl": self.coor_svg2sif(dest_tl),
                "dest_tr": self.coor_svg2sif(dest_tr),
                "dest_br": self.coor_svg2sif(dest_br),
                "dest_bl": self.coor_svg2sif(dest_bl),
            },
        )

        if is_end:
            return layers + [warp]
        else:
            return self.op_encapsulate(layers + [warp])


# ##### Utility Functions ##################################

# ## Path related


def path_to_bline_list(path_d, nodetypes=None, mtx=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]):
    """
    Convert a path to a BLine List

    bline_list format:

    Vertex:
    [[tg1x, tg1y], [x,y], [tg2x, tg2y], split = T/F]
    Vertex list:
    [ vertex, vertex, vertex, ...]
    Bline:
    {
    "points"    : vertex_list,
    "loop"      : True / False
    }
    """

    # Exit on empty paths
    if not path_d:
        return []

    # Parse the path
    path = Path(path_d).to_arrays()

    # Append (more than) enough c's to the nodetypes
    if nodetypes is None:
        nt = ""
    else:
        nt = nodetypes

    for _ in range(len(path)):
        nt += "c"

    # Create bline list
    #     borrows code from cubicsuperpath.py

    # bline_list := [bline, bline, ...]
    # bline := {
    #           "points":[vertex, vertex, ...],
    #           "loop":True/False,
    #          }

    bline_list = []

    subpathstart = []
    last = []
    lastctrl = []
    lastsplit = True

    for s in path:
        cmd, params = s
        if cmd != "M" and bline_list == []:
            raise MalformedSVGError(
                "Bad path data: path doesn't start with moveto, {}, {}".format(s, path)
            )
        elif cmd == "M":
            # Add previous point to subpath
            if last:
                bline_list[-1]["points"].append(
                    [lastctrl[:], last[:], last[:], lastsplit]
                )
            # Start a new subpath
            bline_list.append({"nodetypes": "", "loop": False, "points": []})
            # Save coordinates of this point
            subpathstart = params[:]
            last = params[:]
            lastctrl = params[:]
            lastsplit = False if nt[0] == "z" else True
            nt = nt[1:]
        elif cmd in "LHV":
            bline_list[-1]["points"].append([lastctrl[:], last[:], last[:], lastsplit])
            if cmd == "H":
                last = [params[0], last[1]]
                lastctrl = [params[0], last[1]]
            elif cmd == "V":
                last = [last[0], params[0]]
                lastctrl = [last[0], params[0]]
            else:
                last = params[:]
                lastctrl = params[:]
            lastsplit = False if nt[0] == "z" else True
            nt = nt[1:]
        elif cmd == "C":
            bline_list[-1]["points"].append(
                [lastctrl[:], last[:], params[:2], lastsplit]
            )
            last = params[-2:]
            lastctrl = params[2:4]
            lastsplit = False if nt[0] == "z" else True
            nt = nt[1:]
        elif cmd == "Q":
            q0 = last[:]
            q1 = params[0:2]
            q2 = params[2:4]
            x0 = q0[0]
            x1 = 1.0 / 3 * q0[0] + 2.0 / 3 * q1[0]
            x2 = 2.0 / 3 * q1[0] + 1.0 / 3 * q2[0]
            x3 = q2[0]
            y0 = q0[1]
            y1 = 1.0 / 3 * q0[1] + 2.0 / 3 * q1[1]
            y2 = 2.0 / 3 * q1[1] + 1.0 / 3 * q2[1]
            y3 = q2[1]
            bline_list[-1]["points"].append(
                [lastctrl[:], [x0, y0], [x1, y1], lastsplit]
            )
            last = [x3, y3]
            lastctrl = [x2, y2]
            lastsplit = False if nt[0] == "z" else True
            nt = nt[1:]
        elif cmd == "A":
            from inkex.paths import arc_to_path

            arcp = arc_to_path(last[:], params[:])
            arcp[0][0] = lastctrl[:]
            last = arcp[-1][1]
            lastctrl = arcp[-1][0]
            lastsplit = False if nt[0] == "z" else True
            nt = nt[1:]
            for el in arcp[:-1]:
                el.append(True)
                bline_list[-1]["points"].append(el)
        elif cmd == "Z":
            if len(bline_list[-1]["points"]) == 0:
                # If the path "loops" after only one point
                #  e.g. "M 0 0 Z"
                bline_list[-1]["points"].append([lastctrl[:], last[:], last[:], False])
            elif last == subpathstart:
                # If we are back to the original position
                # merge our tangent into the first point
                bline_list[-1]["points"][0][0] = lastctrl[:]
            else:
                # Otherwise draw a line to the starting point
                bline_list[-1]["points"].append(
                    [lastctrl[:], last[:], last[:], lastsplit]
                )

            # Clear the variables (no more points need to be added)
            last = []
            lastctrl = []
            lastsplit = True

            # Loop the subpath
            bline_list[-1]["loop"] = True
    # Append final superpoint, if needed
    if last:
        bline_list[-1]["points"].append([lastctrl[:], last[:], last[:], lastsplit])

    # Apply the transformation
    if mtx != [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]:
        for bline in bline_list:
            for vertex in bline["points"]:
                for point in vertex:
                    if not isinstance(point, bool):
                        pnt = Transform(mtx).apply_to_point(point)
                        point[0], point[1] = pnt[0], pnt[1]

    return bline_list


# ## Style related


def extract_color(style, color_attrib, *opacity_attribs):
    if color_attrib in style.keys():
        if style[color_attrib] == "none":
            return [1, 1, 1, 0]
        c = inkex.Color(style[color_attrib]).to_rgb()
    else:
        c = (0, 0, 0)

    # Convert color scales and adjust gamma
    color = [
        pow(c[0] / 255.0, sif.gamma),
        pow(c[1] / 255.0, sif.gamma),
        pow(c[2] / 255.0, sif.gamma),
        1.0,
    ]

    for opacity in opacity_attribs:
        if opacity in style.keys():
            color[3] *= float(style[opacity])
    return color


def extract_opacity(style, *opacity_attribs):
    ret = 1.0
    for opacity in opacity_attribs:
        if opacity in style.keys():
            ret *= float(style[opacity])
    return ret


def extract_width(style, width_attrib, mtx):
    if width_attrib in style.keys():
        width = get_dimension(style[width_attrib])
    else:
        width = 1

    area_scale_factor = mtx[0][0] * mtx[1][1] - mtx[0][1] * mtx[1][0]
    linear_scale_factor = math.sqrt(abs(area_scale_factor))

    return width * linear_scale_factor / sif.kux


# ##### Main Class #########################################
class SynfigExport(SynfigPrep):
    def __init__(self):
        SynfigPrep.__init__(self)

    def effect(self):
        # Prepare the document for exporting
        SynfigPrep.effect(self)
        svg = self.document.getroot()
        width = get_dimension(svg.get("width", 1024))
        height = get_dimension(svg.get("height", 768))

        title = svg.getElement("svg:title")
        if title is not None:
            name = title.text
        else:
            name = svg.get("sodipodi:docname", "Synfig Animation 1")

        doc = SynfigDocument(width, height, name)

        layers = []
        for node in svg.iterchildren():
            layers += self.convert_node(node, doc)

        root_canvas = doc.get_root_canvas()
        for layer in layers:
            root_canvas.append(layer)

        self.synfig_document = doc.get_root_tree()

    def save(self, stream):
        self.synfig_document.write(stream)

    def convert_node(self, node, d):
        """Convert an SVG node to a list of Synfig layers"""
        # Parse tags that don't draw any layers
        if isinstance(node, SvgDocumentElement):
            self.parse_defs(node, d)
            return []
        elif not isinstance(
            node, (Group, Anchor, Switch, PathElement, Metadata, NamedView)
        ):
            # An unsupported element
            return []

        layers = []
        if isinstance(node, Group):
            for subnode in node:
                layers += self.convert_node(subnode, d)
            if isinstance(node, Layer):
                name = node.label or "Inline Canvas"
                layers = d.op_encapsulate(layers, name=name)

        elif isinstance(node, (Anchor, Switch)):
            # Treat anchor and switch as a group
            for subnode in node:
                layers += self.convert_node(subnode, d)
        elif isinstance(node, PathElement):
            layers = self.convert_path(node, d)

        style = node.style
        if "filter" in style.keys() and style["filter"].startswith("url"):
            filter_id = style["filter"][5:].split(")")[0]
            layers = d.op_filter(layers, filter_id)

        opacity = extract_opacity(style, "opacity")
        if opacity != 1.0:
            layers = d.op_fade(layers, opacity)

        return layers

    def parse_defs(self, node, d):
        for child in node.iterchildren():
            if isinstance(child, Gradient):
                self.parse_gradient(child, d)
            elif child.TAG == "filter":
                self.parse_filter(child, d)

    def parse_gradient(self, node, d):
        if node.TAG == "linearGradient":
            gradient_id = node.get("id", str(id(node)))
            x1 = float(node.get("x1", "0.0"))
            x2 = float(node.get("x2", "0.0"))
            y1 = float(node.get("y1", "0.0"))
            y2 = float(node.get("y2", "0.0"))

            mtx = node.gradientTransform.matrix

            link = node.get("xlink:href", "#")[1:]
            spread_method = node.get("spreadMethod", "pad")
            if link == "":
                stops = self.parse_stops(node, d)
                d.add_linear_gradient(
                    gradient_id,
                    [x1, y1],
                    [x2, y2],
                    mtx,
                    stops=stops,
                    spread_method=spread_method,
                )
            else:
                d.add_linear_gradient(
                    gradient_id,
                    [x1, y1],
                    [x2, y2],
                    mtx,
                    link=link,
                    spread_method=spread_method,
                )
        elif node.TAG == "radialGradient":
            gradient_id = node.get("id", str(id(node)))
            cx = float(node.get("cx", "0.0"))
            cy = float(node.get("cy", "0.0"))
            r = float(node.get("r", "0.0"))
            fx = float(node.get("fx", "0.0"))
            fy = float(node.get("fy", "0.0"))

            mtx = node.gradientTransform.matrix

            link = node.get("xlink:href", "#")[1:]
            spread_method = node.get("spreadMethod", "pad")
            if link == "":
                stops = self.parse_stops(node, d)
                d.add_radial_gradient(
                    gradient_id,
                    [cx, cy],
                    r,
                    [fx, fy],
                    mtx,
                    stops=stops,
                    spread_method=spread_method,
                )
            else:
                d.add_radial_gradient(
                    gradient_id,
                    [cx, cy],
                    r,
                    [fx, fy],
                    mtx,
                    link=link,
                    spread_method=spread_method,
                )

    def parse_stops(self, node, d):
        stops = {}
        for stop in node.iterchildren():
            if stop.TAG == "stop":
                offset = float(stop.get("offset"))
                style = stop.style
                stops[offset] = extract_color(style, "stop-color", "stop-opacity")
            else:
                raise MalformedSVGError("Child of gradient is not a stop")

        return stops

    def parse_filter(self, node, d):
        filter_id = node.get("id", str(id(node)))

        # A filter is just like an operator (the op_* functions),
        # except that it's created here
        def the_filter(d, layers, is_end=False):
            refs = {None: layers, "SourceGraphic": layers}  # default
            encapsulate_result = not is_end

            for child in node.iterchildren():
                if child.get("in") not in refs:
                    # "SourceAlpha", "BackgroundImage",
                    # "BackgroundAlpha", "FillPaint", "StrokePaint"
                    # are not supported
                    raise UnsupportedException
                l_in = refs[child.get("in")]
                l_out = []
                if child.TAG == "feGaussianBlur":
                    std_dev = child.get("stdDeviation", "0")
                    std_dev = std_dev.replace(",", " ").split()
                    x = float(std_dev[0])
                    if len(std_dev) > 1:
                        y = float(std_dev[1])
                    else:
                        y = x

                    if x == 0 and y == 0:
                        l_out = l_in
                    else:
                        x = d.distance_svg2sif(x)
                        y = d.distance_svg2sif(y)
                        l_out = d.op_blur(l_in, x, y, is_end=True)
                elif child.TAG == "feBlend":
                    # Note: Blend methods are not an exact match
                    # because SVG uses alpha channel in places where
                    # Synfig does not
                    mode = child.get("mode", "normal")
                    if mode == "normal":
                        blend_method = "composite"
                    elif mode == "multiply":
                        blend_method = "multiply"
                    elif mode == "screen":
                        blend_method = "screen"
                    elif mode == "darken":
                        blend_method = "darken"
                    elif mode == "lighten":
                        blend_method = "brighten"
                    else:
                        raise MalformedSVGError("Invalid blend method")

                    if child.get("in2") == "BackgroundImage":
                        encapsulate_result = False
                        l_out = d.op_set_blend(l_in, blend_method) + d.op_set_blend(
                            l_in, "behind"
                        )
                    elif child.get("in2") not in refs:
                        raise UnsupportedException
                    else:
                        l_in2 = refs[child.get("in2")]
                        l_out = l_in2 + d.op_set_blend(l_in, blend_method)

                else:
                    # This filter element is currently unsupported
                    raise UnsupportedException

                # Output the layers
                if child.get("result"):
                    refs[child.get("result")] = l_out

                # Set the default for the next filter element
                refs[None] = l_out

            # Return the output from the last element
            if len(refs[None]) > 1 and encapsulate_result:
                return d.op_encapsulate(refs[None])
            else:
                return refs[None]

        d.add_filter(filter_id, the_filter)

    def convert_path(self, node, d):
        """Convert an SVG path node to a list of Synfig layers"""
        layers = []

        node_id = node.get("id", str(id(node)))
        style = node.style

        mtx = node.transform.matrix
        blines = path_to_bline_list(node.get("d"), node.get("sodipodi:nodetypes"), mtx)
        for bline in blines:
            d.bline_coor_svg2sif(bline)
            bline_guid = d.new_guid()

            if style.setdefault("fill", "#000000") != "none":
                if style["fill"].startswith("url"):
                    # Set the color to black, so we can later overlay
                    # the shape with a gradient or pattern
                    color = [0, 0, 0, 1]
                else:
                    color = extract_color(style, "fill", "fill-opacity")

                layer = d.create_layer(
                    "region",
                    node_id,
                    {
                        "bline": bline,
                        "color": color,
                        "winding_style": 1
                        if style.setdefault("fill-rule", "nonzero") == "evenodd"
                        else 0,
                    },
                    guids={"bline": bline_guid},
                )

                if style["fill"].startswith("url"):
                    color_layer = self.convert_url(
                        style["fill"][5:].split(")")[0], mtx, d
                    )[0]
                    layer = d.op_color([layer], overlay=color_layer)[0]
                    layer = d.op_fade([layer], extract_opacity(style, "fill-opacity"))[
                        0
                    ]

                layers.append(layer)

            if style.setdefault("stroke", "none") != "none":
                if style["stroke"].startswith("url"):
                    # Set the color to black, so we can later overlay
                    # the shape with a gradient or pattern
                    color = [0, 0, 0, 1]
                else:
                    color = extract_color(style, "stroke", "stroke-opacity")

                layer = d.create_layer(
                    "outline",
                    node_id,
                    {
                        "bline": bline,
                        "color": color,
                        "width": extract_width(style, "stroke-width", mtx),
                        "sharp_cusps": True
                        if style.setdefault("stroke-linejoin", "miter") == "miter"
                        else False,
                        "round_tip[0]": False
                        if style.setdefault("stroke-linecap", "butt") == "butt"
                        else True,
                        "round_tip[1]": False
                        if style.setdefault("stroke-linecap", "butt") == "butt"
                        else True,
                    },
                    guids={"bline": bline_guid},
                )

                if style["stroke"].startswith("url"):
                    color_layer = self.convert_url(
                        style["stroke"][5:].split(")")[0], mtx, d
                    )[0]
                    layer = d.op_color([layer], overlay=color_layer)[0]
                    layer = d.op_fade(
                        [layer], extract_opacity(style, "stroke-opacity")
                    )[0]

                layers.append(layer)

        return layers

    def convert_url(self, url_id, mtx, d):
        """Return a list Synfig layers that represent the gradient with the given id"""
        gradient = d.get_gradient(url_id)
        if gradient is None:
            # Patterns and other URLs not supported
            return [None]

        if gradient["type"] == "linear":
            layer = d.create_layer(
                "linear_gradient",
                url_id,
                d.gradient_to_params(gradient),
                guids={"gradient": gradient["stops_guid"]},
            )

        if gradient["type"] == "radial":
            layer = d.create_layer(
                "radial_gradient",
                url_id,
                d.gradient_to_params(gradient),
                guids={"gradient": gradient["stops_guid"]},
            )

        trm = Transform(mtx) * Transform(gradient["mtx"])
        return d.op_transform([layer], trm.matrix)


if __name__ == "__main__":
    SynfigExport().run()



#!/usr/bin/env python
#
# Copyright (C) 2010 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Common elements between webslicer extensions
"""

import inkex
from inkex import Group


def is_empty(val):
    return val in ("", None)


class WebSlicerMixin(object):
    def get_slicer_layer(self, force_creation=False):
        # Test if webslicer-layer layer existis
        layer = self.svg.getElement(
            '//*[@id="webslicer-layer" and @inkscape:groupmode="layer"]'
        )
        if layer is None:
            if force_creation:
                # Create a new layer
                layer = Group(id="webslicer-layer")
                layer.set("inkscape:label", "Web Slicer")
                layer.set("inkscape:groupmode", "layer")
                self.document.getroot().append(layer)
            else:
                layer = None
        return layer

    def get_conf_text_from_list(self, conf_atts):
        conf_list = []
        for att in conf_atts:
            if not is_empty(getattr(self.options, att)):
                conf_list.append(
                    att.replace("_", "-") + ": " + str(getattr(self.options, att))
                )
        return "\n".join(conf_list)



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2021 Windell H. Oskay, www.evilmadscientist.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
An Inkscape extension to assist with importing AI SVG files.
"""

import inkex
from inkex import units


class DocAiConvert(inkex.EffectExtension):
    """
    Main class of the doc_ai_convert extension

    This extension performs the following two actions:
    1) Recognize intended dimensions of original document.
        - If the AI document was set up with a size given in picas, cm, mm, or in,
          and then exported to SVG, no conversion is needed.
        - If the AI document was set up with any other dimensions,
          (point, yard, px, ft/in, ft, or m), then the exported SVG
          will be in units of 72 DPI pixels. This causes an issue,
          as Inkscape uses CSS pixels (px) which are at 96 DPI.
          In this case, we re-interpret the size as 72 DPI points (pt),
          which will convert the artwork to appear at the correct size.
    2) Recognize Adobe Illustrator layers.
        - Group (<g>) elements with non-empty data-name attributes are
          groups in an Illustrator SVG document.
        - We re-label these as layers such that Inkscape will recognize them.

    """

    def effect(self):
        """
        Main entry point of the doc_ai_convert extension
        """
        # 1) Recognize intended dimensions of original document, if given
        width_string = self.svg.get("width")
        height_string = self.svg.get("height")

        if width_string and height_string:
            width_num, width_units = units.parse_unit(width_string)
            height_num, height_units = units.parse_unit(height_string)

            # Note that units.parse_unit will return units of 'px'
            #    for unitless values, and not None.

            if width_num:
                if width_units == "px":
                    self.svg.set("width", units.render_unit(width_num, "pt"))
            if height_num:
                if height_units == "px":
                    self.svg.set("height", units.render_unit(height_num, "pt"))

        # 2) Recognize Adobe Illustrator layers.
        for node in self.svg.xpath("//svg:g[@data-name]"):
            node.set("inkscape:groupmode", "layer")
            node.set("inkscape:label", node.pop("data-name"))


if __name__ == "__main__":
    DocAiConvert().run()



# coding=utf-8
#
# Copyright (C) 2010 Geoffrey Mosini
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Generate barcodes for Code25-interleaved 2 of 5, for Inkscape.
"""

from .Base import Barcode
from inkex.localization import inkex_gettext as _

# 1 means thick, 0 means thin
ENCODE = {
    "0": "00110",
    "1": "10001",
    "2": "01001",
    "3": "11000",
    "4": "00101",
    "5": "10100",
    "6": "01100",
    "7": "00011",
    "8": "10010",
    "9": "01010",
}


class Code25i(Barcode):
    """Convert a text into string binary of black and white markers"""

    # Start and stop code are already encoded into white (0) and black(1) bars
    def encode(self, text):
        if not text.isdigit():
            return self.error(text, _("CODE25 can only encode numbers."))
        number = text

        # Number of figures to encode must be even,
        # a 0 is added to the left in case it's odd.
        if len(number) % 2 > 0:
            number = "0" + number

        # Number is encoded by pairs of 2 figures
        size = len(number) // 2
        encoded = "1010"
        for i in range(size):
            # First in the pair is encoded in black (1), second in white (0)
            black = ENCODE[number[i * 2]]
            white = ENCODE[number[i * 2 + 1]]
            for j in range(5):
                if black[j] == "1":
                    encoded += "11"
                else:
                    encoded += "1"
                if white[j] == "1":
                    encoded += "00"
                else:
                    encoded += "0"
        return encoded + "1101"



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
# Copyright 2022 Simon Duerr <dev@simonduerr.eu>
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from import_sources import RemoteSource


license_mapping = {
    "PD": "cc-0",
    "CC-0": "cc-0",
    "CC-BY 3.0": "cc-by-3.0",
    "CC-BY 4.0": "cc-by-4.0",
    "CC-BY-SA 4.0": "cc-by-sa-4.0",
    "CC-BY-NC 3.0": "cc-by-nc-3.0",
    "CC-BY-NC-SA 3.0": "cc-by-nc-sa-3.0",
    "CC-BY-NC-SA 4.0": "cc-by-nc-sa-4.0",
    "CC-BY-SA 3.0": "cc-by-sa-3.0",
    "CC-BY-ND 3.0": "cc-by-nd-3.0",
    "ASL": "asl",
    "nbsd": "bsd",
    "GPLv2": "gpl-2",
    "GPLv3": "gpl-3",
    "AGPLv3": "agpl-3",
    "MIT": "mit",
}


class InkscapeWebsite(RemoteSource):
    name = "Inkscape Community"
    icon = "sources/inkscape-web.svg"
    is_default = True

    base_url = "https://inkscape.org/gallery/=artwork/json/"

    def search(self, query):
        """Ask the inkscape website for some artwork"""
        items = []
        try:
            response = self.session.get(self.base_url, params={"q": query})
            items = response.json()["items"]
        except Exception:
            pass

        for item in items:
            if "svg" not in item["type"]:
                continue
            if (
                item["license"] == "(C)"
            ):  # ignore copyrighted items because they cannot be reused
                continue
            yield {
                "id": item["id"],
                "name": item["name"],
                "author": item["author"],
                "license": license_mapping.get(item["license"], item["license"].lower()),
                "summary": item["summary"],
                "thumbnail": item["icon"] or item["links"]["file"],
                "created": item["dates"]["created"],
                "popularity": item["stats"]["liked"],
                "file": item["links"]["file"],
            }
